""" PlotItem with matlab color scheme and CustomViewBox.
"""

from __future__ import annotations
from qtpy.QtGui import QColor
import pyqtgraph as pg
from pyqt_ext.pyqtgraph_ext import View


class Plot(pg.PlotItem):
    """ PlotItem with matlab color scheme and custom ViewBox.
    
    !!! If you provide a viewBox during instantiation,
        it will be replaced with a new ViewBox instance.
        To get the valid viewBox reference after plot creation,
        use `self.getViewBox()`.
    """

    def __init__(self, *args, **kwargs):
        if 'viewBox' not in kwargs:
            kwargs['viewBox'] = View()
        if 'pen' not in kwargs:
            # MATLAB color scheme
            kwargs['pen'] = pg.mkPen(QColor.fromRgbF(0.15, 0.15, 0.15), width=1)
        pg.PlotItem.__init__(self, *args, **kwargs)

        # MATLAB color scheme
        for axis in ['left', 'bottom', 'right', 'top']:
            axis_item = self.getAxis(axis)
            if axis_item is not None:
                axis_item.setTextPen(QColor.fromRgbF(0.15, 0.15, 0.15))



""" PySide/PyQt color utils.
"""

from qtpy.QtGui import QColor


ColorType = str | tuple[int | float] | list[int | float] | QColor | None


def toQColor(color: ColorType, name_map: dict[str, ColorType] = None) -> QColor:
    """ Convert a color object to a QColor.
    """
    if color is None:
        return QColor('transparent')
    if isinstance(color, QColor):
        return color
    if isinstance(color, str):
        color = color.strip()
        if (name_map is not None) and (color in name_map):
            return toQColor(name_map[color])
        elif QColor.isValidColorName(color):
            return QColor(color)
        elif color.lower() == 'none':
            return QColor('transparent')
        else:
            # (r,g,b) or (r,g,b,a)
            color = color.lstrip('(').rstrip(')').split(',')
            try:
                color = [int(part) for part in color]
            except:
                color = [float(part) for part in color]
    # (r,g,b) or (r,g,b,a)
    if isinstance(color[0], int):
        return QColor(*color)
    elif isinstance(color[0], float):
        return QColor.fromRgbF(*color)


def toColorStr(color: ColorType) -> str:
    """ Convert a color object to a string representation of the color.
    """
    if color is None:
        return 'none'
    if isinstance(color, QColor):
        # (r,g,b,a) in [0,255]
        return f'({color.red()}, {color.green()}, {color.blue()}, {color.alpha()})'
    if isinstance(color, str):
        if ',' in color:
            # (r,g,b) or (r,g,b,a)
            return toColorStr(toQColor(color))
        # assume color str is valid 
        return color
    if isinstance(color, tuple) or isinstance(color, list):
        # (r,g,b) or (r,g,b,a)
        return '(' + ', '.join([str(part) for part in color]) + ')'


from pyqt_ext.widgets.ColorButton import ColorButton
from pyqt_ext.widgets.CollapsibleSection import CollapsibleSection
from pyqt_ext.widgets.FlowLayout import FlowLayout
from pyqt_ext.widgets.MultiValueSpinBox import MultiValueSpinBox
from pyqt_ext.widgets.TableWidgetWithCopyPaste import TableWidgetWithCopyPaste


""" ViewBox with matlab color scheme and context menu for drawing ROIs and events.
"""

from __future__ import annotations
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
import numpy as np
import pyqtgraph as pg
from pyqt_ext.pyqtgraph_ext import XAxisRegion, YAxisRegion, Graph


class View(pg.ViewBox):
    """ ViewBox with functionality for drawing ROIs.
    """

    sigStartedDrawingItems = Signal()
    sigItemAdded = Signal(QGraphicsObject)  # emits the newly added QGraphicsObject item
    sigFinishedDrawingItems = Signal()

    def __init__(self, *args, **kwargs):
        pg.ViewBox.__init__(self, *args, **kwargs)

        self._lastMousePressPosInAxesCoords = {}  # dict keys are mouse buttons
        self._drawingItemsOfType = None
        self._itemBeingDrawn = None

        # MATLAB color scheme
        self.setBackgroundColor(QColor(255, 255, 255))

        # colormap (MATLAB lines)
        # rows are colors
        self._colormap = [
            (  0, 114, 189),
            (217,  83,  25),
            (237, 177,  32),
            (126,  47, 142),
            (119, 172,  48),
            ( 77, 190, 238),
            (162,  20,  47),
        ]

        # selected color (row) in colormap
        self._colorIndex = 0

        # ROI styles
        self._ROI_pen = pg.mkPen(QColor(237, 135, 131), width=1)
        self._ROI_hoverPen = pg.mkPen(QColor(255, 0, 0), width=2)
        self._ROI_handlePen = self._ROI_pen
        self._ROI_handleHoverPen = self._ROI_hoverPen
        self._ROI_brush = pg.mkBrush(QColor(237, 135, 131, 51))
        self._ROI_hoverBrush = pg.mkBrush(QColor(237, 135, 131, 128))
    
    def colormap(self):
        return self._colormap
    
    def setColormap(self, colormap) -> None:
        self._colormap = colormap
        self._colorIndex = self._colorIndex % len(self._colormap)
    
    def colorAtIndex(self, colorIndex) -> QColor:
        ncolors = len(self._colormap)
        color = self._colormap[colorIndex % ncolors]
        return QColor(*color)
    
    def nextColor(self) -> QColor:
        ncolors = len(self._colormap)
        color = self._colormap[self._colorIndex % ncolors]
        self._colorIndex = (self._colorIndex + 1) % ncolors
        return QColor(*color)
    
    def colorIndex(self) -> int:
        return self._colorIndex
    
    def setColorIndex(self, colorIndex: int) -> None:
        ncolors = len(self._colormap)
        self._colorIndex = colorIndex % ncolors
    
    def mousePressEvent(self, event):
        # store mouse press position in axes coords
        posInAxesCoords = self.mapSceneToView(self.mapToScene(event.pos()))
        self._lastMousePressPosInAxesCoords[event.button()] = posInAxesCoords

        if event.button() == Qt.LeftButton:
            # start drawing a new item?
            newItem = None
            if self._drawingItemsOfType is not None:
                if self._drawingItemsOfType == XAxisRegion:
                    limits = [posInAxesCoords.x(), posInAxesCoords.x()]
                    newItem = XAxisRegion(values=limits, pen=self._ROI_pen, hoverPen=self._ROI_hoverPen, brush=self._ROI_brush, hoverBrush=self._ROI_hoverBrush)
                elif self._drawingItemsOfType == YAxisRegion:
                    limits = [posInAxesCoords.y(), posInAxesCoords.y()]
                    newItem = YAxisRegion(values=limits, pen=self._ROI_pen, hoverPen=self._ROI_hoverPen, brush=self._ROI_brush, hoverBrush=self._ROI_hoverBrush)
                elif self._drawingItemsOfType in [pg.RectROI, pg.EllipseROI, pg.CircleROI, pg.LineSegmentROI]:
                    newItem = self._drawingItemsOfType(pos=posInAxesCoords, size=[0, 0], invertible=True, pen=self._ROI_pen, hoverPen=self._ROI_hoverPen, handlePen=self._ROI_handlePen, handleHoverPen=self._ROI_handleHoverPen)
                elif issubclass(self._drawingItemsOfType, pg.PlotDataItem):
                    if isinstance(self._itemBeingDrawn, pg.PlotDataItem):
                        # add point to existing Graph
                        x, y = self._itemBeingDrawn.getOriginalDataset()
                        x = np.append(x, posInAxesCoords.x())
                        y = np.append(y, posInAxesCoords.y())
                        self._itemBeingDrawn.setData(x, y)
                        event.accept()
                        return
                    else:
                        newItem = Graph(pen=self._ROI_pen, symbol='o', symbolPen=self._ROI_pen, symbolBrush=self._ROI_brush)
                        newItem.setData([posInAxesCoords.x()], [posInAxesCoords.y()])
                if newItem is not None:
                    self._itemBeingDrawn = newItem
                    self.addItem(self._itemBeingDrawn)
                    event.accept()
                    return
        
        # default if event was not handled above
        pg.ViewBox.mousePressEvent(self, event)
    
    def mouseReleaseEvent(self, event):
        if event.button() == Qt.LeftButton:
            # finished drawing region/event?
            if  self._itemBeingDrawn is not None:
                if type(self._itemBeingDrawn) in [XAxisRegion, YAxisRegion, pg.RectROI, pg.EllipseROI, pg.CircleROI, pg.LineSegmentROI]:
                    self.sigItemAdded.emit(self._itemBeingDrawn)
                    self._itemBeingDrawn = None
                event.accept()
                return
        
        # default if event was not handled above
        pg.ViewBox.mouseReleaseEvent(self, event)
    
    def mouseMoveEvent(self, event):
        if event.buttons() & Qt.LeftButton:
            # drawing region?
            if self._itemBeingDrawn is not None:
                startPosInAxesCoords = self._lastMousePressPosInAxesCoords[Qt.LeftButton]
                posInAxesCoords = self.mapSceneToView(self.mapToScene(event.pos()))
                if isinstance(self._itemBeingDrawn, XAxisRegion):
                    limits = sorted([startPosInAxesCoords.x(), posInAxesCoords.x()])
                    self._itemBeingDrawn.setRegion(limits)
                elif isinstance(self._itemBeingDrawn, YAxisRegion):
                    limits = sorted([startPosInAxesCoords.y(), posInAxesCoords.y()])
                    self._itemBeingDrawn.setRegion(limits)
                elif type(self._itemBeingDrawn) in [pg.RectROI, pg.EllipseROI, pg.CircleROI]:
                    self._itemBeingDrawn.setSize(posInAxesCoords - self._itemBeingDrawn.pos())
                elif isinstance(self._itemBeingDrawn, pg.LineSegmentROI):
                    state = self._itemBeingDrawn.getState()
                    state['points'] = [pg.Point(startPosInAxesCoords), pg.Point(posInAxesCoords)]
                    self._itemBeingDrawn.setState(state)
                event.accept()
                return
        
        # default if event was not handled above
        pg.ViewBox.mouseMoveEvent(self, event)
    
    def startDrawingItemsOfType(self, itemType):
        self._itemBeingDrawn = None
        self._drawingItemsOfType = itemType
        self.sigStartedDrawingItems.emit()
    
    def stopDrawingItems(self):
        self._drawingItemsOfType = None
        self._itemBeingDrawn = None
        self.sigFinishedDrawingItems.emit()
    
    def listItemsOfType(self, itemType):
        return [item for item in self.allChildren() if isinstance(item, itemType)]


def test_live():
    import numpy as np
    from pyqt_ext.pyqtgraph_ext import Figure, Graph
    app = QApplication()

    plot = Figure(viewBox=View())
    # !!! MUST get view ref after plot creation, WTF!?
    view = plot.getViewBox()
    line = Graph(y=np.random.randn(1000))
    plot.addItem(line)
    plot.setWindowTitle('pyqtgraph-tools')
    plot.show()

    view.startDrawingItemsOfType(Graph)
    QTimer.singleShot(2500, lambda: view.stopDrawingItems())
    QTimer.singleShot(2600, lambda: view.startDrawingItemsOfType(XAxisRegion))
    QTimer.singleShot(5000, lambda: view.stopDrawingItems())

    app.exec()


if __name__ == '__main__':
    test_live()



""" Data interface for a tree of key:value pairs.

e.g., Any level of nesting of dict and list objects.

TODO:
- Numpy N-D arrays
- nested lists/tuples
"""

from __future__ import annotations
from warnings import warn
from typing import Any
import numpy as np
from pyqt_ext.tree import AbstractTreeItem


class KeyValueTreeItem(AbstractTreeItem):
    
    def __init__(self, key: str | None = None, value: Any = None, parent: KeyValueTreeItem | None = None) -> None:
        self._key = key
        self._value = value
        AbstractTreeItem.__init__(self, parent=parent)
        self.setupSubtree()
    
    def key(self) -> str | int | None:
        # if parent is a list, return this item's sibling index
        parent: KeyValueTreeItem = self.parent()
        if parent is not None:
            parent_map: dict | list = parent.value()
            if isinstance(parent_map, list):
                return self.siblingIndex()
        
        # otherwise, return the stored key
        return self._key
    
    def setKey(self, key: str) -> None:
        # if parent is a dict, update the parent dict to reflect the key change
        parent: KeyValueTreeItem = self.parent()
        if parent is not None:
            parent_map: dict | list = parent.value()
            if isinstance(parent_map, dict):
                # must have a valid key
                if key is None:
                    key = 'None'
                elif key == '':
                    return
                elif type(key) not in [str, int, float, bool]:
                    return
                # check if key has changed
                if key == self._key:
                    return
                # make sure key is unique
                if key in parent_map:
                    warn(f'Key {key} already exists.')
                    # key = self.uniqueName(key, list(parent_map.keys()))
                    return
                # swap keys in map
                parent_map[key] = parent_map.pop(self._key)
        
        # set the stored key
        self._key = key
    
    def value(self) -> Any:
        return self._value
    
    def setValue(self, value: Any) -> None:
        # print(value, type(value), flush=True)
        if isinstance(self._value, np.ndarray) and (isinstance(value, list) or isinstance(value, tuple)):
            value = np.array(value)
        
        self._value = value

        # update the parent key:value map to reflect the value change
        parent: KeyValueTreeItem = self.parent()
        if parent is not None:
            parent_map: dict | list = parent.value()
            parent_map[self.key()] = value
        
        # reset this item's subtree
        for child in self.children:
            child._parent = None
        self.children = []
        self.setupSubtree()
    
    def setupSubtree(self):
        """ Recursively build subtree if value is itself a container with key:value access.
        """
        value = self.value()
        if isinstance(value, dict):
            for key, val in value.items():
                KeyValueTreeItem(key, val, parent=self)
        elif isinstance(value, list):
            for val in value:
                # list keys are not explicitly set, they will default to the list index
                KeyValueTreeItem(None, val, parent=self)
    
    def name(self) -> str:
        """ name <==> key
        """
        key = self.key()
        return str(key) if key is not None else None
    
    def setName(self, name: str) -> None:
        """ name <==> key
        """
        self.setKey(name)
    
    def __repr__(self):
        key = self.key()
        value = self.value()
        if isinstance(value, dict):
            return f'{key}: {{}}'
        elif isinstance(value, list):
            return f'{key}: []'
        else:
            return f'{key}: {value} {type(value)}'
    
    def __str__(self) -> str:
        return self._tree_repr(lambda item: item.__repr__())
    
    def setParent(self, parent: KeyValueTreeItem | None) -> None:
        old_parent: KeyValueTreeItem | None = self.parent()
        new_parent: KeyValueTreeItem | None = parent
        if old_parent is new_parent:
            # nothing to do
            return
        if new_parent is not None:
            if new_parent.hasAncestor(self):
                raise ValueError('Cannot set parent to a descendant.')
            # the new parent must be a key:value container
            new_parent_value = new_parent.value()
            if not isinstance(new_parent_value, dict) and not isinstance(new_parent_value, list):
                raise ValueError('Parent must be a key:value mapping (dict or list).')
        
        if old_parent is not None:
            # remove from old parent mapping
            old_parent_map: dict | list = old_parent.value()
            # detach from old parent item
            old_parent_map.pop(self.key())
            old_parent.children.remove(self)
        
        if new_parent is not None:
            # attach to new parent item (appends as last child)
            new_parent.children.append(self)
            self._parent = new_parent
            # add to new parent mapping (unless it already exists such as when building an item tree for an existing data tree)
            new_parent_map: dict | list = new_parent.value()
            if isinstance(new_parent_map, dict):
                key = self.key()
                value = self.value()
                # ensure valid dict key
                if (key is None) or (key == ''):
                    sibling_keys = [sibling.key() for sibling in self.siblings()]
                    key = self.uniqueName('?', sibling_keys)
                if key not in new_parent_map:
                    new_parent_map[key] = value
                    self._key = key
                else:
                    # data at key already exists
                    # if it's the same as this item's value, do nothing -> building an item tree for an existing data tree
                    # otherwise, update the parent map's value
                    existing_value = new_parent_map[key]
                    if isinstance(existing_value, dict) or isinstance(existing_value, list):
                        if existing_value is not value:
                            self.setValue(value)
                    elif isinstance(existing_value, np.ndarray) or isinstance(value, np.ndarray):
                        if np.issubdtype(existing_value.dtype, np.floating):
                            if not np.allclose(value, existing_value):
                                self.setValue(value)
                        else:
                            if not np.array_equal(value, existing_value):
                                self.setValue(value)
                    else:
                        if existing_value != value:
                            self.setValue(value)
            elif isinstance(new_parent_map, list):
                index: int = self.siblingIndex()
                value = self.value()
                while len(new_parent_map) < index:
                    # should not happen
                    new_parent_map.append(None)
                if len(new_parent_map) == index:
                    new_parent_map.append(value)
                else:
                    # data at index already exists
                    # if it's the same as this item's value, do nothing -> building an item tree for an existing data tree
                    # otherwise, update the parent map's value
                    existing_value = new_parent_map[index]
                    if isinstance(existing_value, dict) or isinstance(existing_value, list):
                        if existing_value is not value:
                            self.setValue(value)
                    elif isinstance(existing_value, np.ndarray) or isinstance(value, np.ndarray):
                        if np.issubdtype(existing_value.dtype, np.floating):
                            if not np.allclose(value, existing_value):
                                self.setValue(value)
                        else:
                            if not np.array_equal(value, existing_value):
                                self.setValue(value)
                    else:
                        if existing_value != value:
                            self.setValue(value)
    
    def insertChild(self, index: int, child: KeyValueTreeItem) -> None:
        if not (0 <= index <= len(self.children)):
            raise IndexError('Index out of range.')
        
        # append as last child
        child.setParent(self)
        
        # move item to index
        pos = self.children.index(child)
        if pos != index:
            if pos < index:
                index -= 1
            if pos != index:
                # reorder key:value map
                this_map: dict | list = self.value()
                if isinstance(this_map, dict):
                    pass  # TODO... reorder dict? Not absolutely necessary.
                elif isinstance(this_map, list):
                    this_map.insert(index, this_map.pop(pos))
                # reorder child items
                self.children.insert(index, self.children.pop(pos))


def test_tree():
    import json
    tree = {
        'a': 1,
        'b': [4, 8, 9, 5, 7, 99],
        'c': {
            'me': 'hi',
            3: 67,
            'd': {
                'e': 3,
                'f': 'ya!',
                'g': 5,
            },
        },
        'nd': np.array([[1, 2, 3], [4, 5, 6]]),
    }
    # print(json.dumps(tree, indent='    '))
    root = KeyValueTreeItem(None, tree)
    print('-'*82)
    print(root)
    # print(json.dumps(tree, indent='    '))

    print('-'*82)
    print('remove /a')
    root.removeChild(root['/a'])
    print(root)

    print('-'*82)
    print('remove /c/d/f')
    d = root['c/d']
    f = root['c/d/f']
    d.removeChild(f)
    print(root)

    print('-'*82)
    print('move /c/d to /d')
    d.setParent(root)
    print(root)

    print('-'*82)
    print('move /d to 2nd child of /')
    root.insertChild(1, d)
    print(root)

    print('-'*82)
    print('move /d to 3rd child of /b')
    b = root['b']
    b.insertChild(2, d)
    print(root)

    print('-'*82)
    print('remove /b/1')
    b.removeChild(b['1'])
    print(root)

    print('-'*82)
    print('/c/me:hi -> /c/me:bye')
    c = root['c']
    c.children[0].setValue('bye')
    print(root)

    print('-'*82)
    print('move /b/1 to first child of /c')
    c.insertChild(0, b['1'])
    print(root)

    print('-'*82)
    print('/c -> 82')
    c.setValue(82)
    print(root)

    print('-'*82)
    print('/c -> {a:1, b:2}')
    c.setValue({'a': 1, 'b': 2})
    print(root)

    print('-'*82)
    print('/nd')
    print(root['/nd'])


if __name__ == '__main__':
    test_tree()



""" PySide/PyQt button for selecting and displaying a color.
"""

from qtpy.QtCore import Signal
from qtpy.QtGui import QColor, QIcon
from qtpy.QtWidgets import QToolButton, QColorDialog
import qtawesome as qta
from pyqt_ext.utils import ColorType, toQColor


class ColorButton(QToolButton):
    """ Button for selecting and displaying a color.
    """

    colorChanged = Signal(QColor)
    
    def __init__(self, color = None):
        QToolButton.__init__(self)
        self.setColor(color)
        self.clicked.connect(self.pickColor)
    
    def color(self) -> QColor | None:
        return self._color

    def setColor(self, color: ColorType):
        if color is None:
            color: QColor = QColor('transparent')
            self.setStyleSheet(f'background-color: rgba({color.red()}, {color.green()}, {color.blue()}, {color.alpha()}); border: 1px solid black;')
            self.setIcon(qta.icon('ri.question-mark'))
            self._color = None
            return
        color: QColor = toQColor(color)
        self.setStyleSheet(f'background-color: rgba({color.red()}, {color.green()}, {color.blue()}, {color.alpha()}); border: 1px solid black;')
        self.setIcon(QIcon())
        self._color = color
        self.colorChanged.emit(color)
    
    def pickColor(self):
        color: QColor = self.color()
        if color is None:
            color = QColor('white')
        color = QColorDialog.getColor(color, self, "Select Color", options=QColorDialog.ShowAlphaChannel)
        if color.isValid():
            self.setColor(color)


def test_live():
    from qtpy.QtWidgets import QApplication, QWidget, QVBoxLayout

    app = QApplication()

    redButton = ColorButton('red')
    transparentGreenButton = ColorButton([0, 255, 0, 64])
    blueButton = ColorButton([0.0, 0.0, 1.0])
    noColorButton = ColorButton()

    redButton.colorChanged.connect(print)
    transparentGreenButton.colorChanged.connect(print)
    blueButton.colorChanged.connect(print)
    noColorButton.colorChanged.connect(print)

    ui = QWidget()
    vbox = QVBoxLayout(ui)
    vbox.addWidget(redButton)
    vbox.addWidget(transparentGreenButton)
    vbox.addWidget(blueButton)
    vbox.addWidget(noColorButton)
    ui.show()

    app.exec()

    print('Final color selections:')
    print(f'red -> {redButton.color()}')
    print(f'transparentGreen -> {transparentGreenButton.color()}')
    print(f'blue -> {blueButton.color()}')
    print(f'noColor -> {noColorButton.color()}')


if __name__ == '__main__':
    test_live()



""" Data interface and widgets for storing/editing the style of a graph.

Style is stored in hashable dict.
"""

from __future__ import annotations
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from pyqt_ext.utils import ColorType, toQColor, toColorStr
from pyqt_ext.widgets import ColorButton, CollapsibleSection


class GraphStyle(dict):
    """ Hashable style dict for graph data.

    'color': str
    'linestyle': str
    'linewidth': float
    'marker': str
    'markersize': float
    'markeredgestyle': str
    'markeredgewidth': float
    'markeredgecolor': str
    'markerfacecolor': str
    """

    # alternate key names
    keymap = {
        'c': 'color',
        'ls': 'linestyle',
        'lw': 'linewidth',
        'symbol': 'marker',
        'm': 'marker',
        'ms': 'markersize',
        'mes': 'markeredgestyle',
        'mew': 'markeredgewidth',
        'mec': 'markeredgecolor',
        'mfc': 'markerfacecolor',
    }

    # lines
    lineStyles = ['none', '-', '--', ':', '-.', '-..']
    penStyles = [Qt.PenStyle.NoPen, Qt.PenStyle.SolidLine, Qt.PenStyle.DashLine, Qt.PenStyle.DotLine, Qt.PenStyle.DashDotLine, Qt.PenStyle.DashDotDotLine]
    penStyleLabels = ['No Line', 'Solid Line', 'Dash Line', 'Dot Line', 'Dash Dot Line', 'Dash Dot Dot Line']

    # markers {label: marker key}
    pyqtgraphMarkers = {
        'None': 'none',
        'Circle': 'o',
        'Square': 's',
        'Triangle': 't',
        'Diamond': 'd',
        'Plus': '+',
        'Triangle Up': 't1',
        'Triangle Right': 't2',
        'Triangle Left': 't3',
        'Pentagon': 'p',
        'Hexagon': 'h',
        'Star': 'star',
        'Vertical Line': '|',
        'Horizontal Line': '_',
        'Cross': 'x',
        'Arrow Up': 'arrow_up',
        'Arrow Right': 'arrow_right',
        'Arrow Down': 'arrow_down',
        'Arrow Left': 'arrow_left',
        'Crosshair': 'crosshair'
    }
    # pyqtgraphMarkers = ['none', 'circle', 'triangle down', 'triangle up', 'triangle right', 'triangle left', 'square', 'diamond', 'pentagon', 'hexagon', 'star', 'plus', 'cross']

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)

        # default values
        self._defaults = {
            'linestyle': '-',
            'linewidth': 1,
            'marker': 'none',
            'markersize': 10,
            'markeredgestyle': '-',
            'markeredgewidth': 1,
        }
    
    def __getitem__(self, key: str):
        key = self.getKey(key)
        if key in self:
            return dict.__getitem__(self, key)
        # key not found...
        if key == 'markeredgewidth':
            if 'linewidth' in self:
                return self['linewidth']
        elif key == 'markeredgecolor':
            if 'color' in self:
                return self['color']
        elif key == 'markerfacecolor':
            if 'markeredgecolor' in self:
                return self['markeredgecolor']
            elif 'color' in self:
                return self['color']
        if key in self._defaults:
            return self._defaults[key]
    
    def __setitem__(self, key: str, value):
        key = self.getKey(key)
        if key.endswith('color'):
            if value is not None:
                value = toColorStr(value)
        elif key in ['linestyle', 'markeredgestyle']:
            if value is None:
                value = 'none'
            elif isinstance(value, int):
                value = GraphStyle.lineStyles[value]
            elif isinstance(value, Qt.PenStyle):
                value = GraphStyle.lineStyles[GraphStyle.penStyles.index(value)]
            elif isinstance(value, str):
                if value == '.-':
                    value = '-.'
                elif value == '..-' or value == '.-.':
                    value = '-..'
        elif key == 'linewidth':
            if value is not None:
                value = max(0, value)
        elif key == 'marker':
            if isinstance(value, str) and value.lower() == 'none':
                value = None
            # elif value in GraphStyle.pyqtgraphMarkers.values():
            #     index = list(GraphStyle.pyqtgraphMarkers.values()).index(value)
            #     value = list(GraphStyle.pyqtgraphMarkers.keys())[index]
        elif key == 'markersize':
            if value is not None:
                value = max(0, value)
        elif key == 'markeredgewidth':
            if value is not None:
                value = max(0, value)
        if value is None:
            del self[key]
            return
        dict.__setitem__(self, key, value)
    
    def __delitem__(self, key: str):
        key = self.getKey(key)
        if key in self:
            dict.__delitem__(self, key)
    
    @staticmethod
    def getKey(key: str) -> str:
        key = key.lower()
        if key in GraphStyle.keymap:
            key = GraphStyle.keymap[key]
        return key

    def createWidget(self, key: str) -> QWidget:
        key = self.getKey(key)
        if key in ['color', 'markeredgecolor', 'markerfacecolor']:
            widget = ColorButton()
            widget.setColor(self[key])
            return widget
        elif key in ['linestyle', 'markeredgestyle']:
            widget = QComboBox()
            widget.addItems(GraphStyle.penStyleLabels)
            widget.setCurrentIndex(GraphStyle.lineStyles.index(self[key]))
            return widget
        elif key in ['linewidth', 'markersize', 'markeredgewidth']:
            widget = QDoubleSpinBox()
            widget.setMinimum(0)
            widget.setValue(self[key])
            return widget
        elif key == 'marker':
            widget = QComboBox()
            marker_labels = list(GraphStyle.pyqtgraphMarkers.keys())
            marker_keys = list(GraphStyle.pyqtgraphMarkers.values())
            widget.addItems(marker_labels)
            marker = self[key]
            if marker is None:
                marker = 'None'
            elif marker in marker_keys:
                marker = marker_labels[marker_keys.index(marker)]
            widget.setCurrentIndex(marker_labels.index(marker))
            return widget

    def updateWidget(self, key: str, widget: QWidget) -> None:
        key = self.getKey(key)
        if key in ['color', 'markeredgecolor', 'markerfacecolor']:
            widget.setColor(self[key])
        elif key in ['linestyle', 'markeredgestyle']:
            widget.setCurrentIndex(GraphStyle.lineStyles.index(self[key]))
        elif key in ['linewidth', 'markersize', 'markeredgewidth']:
            widget.setValue(self[key])
        elif key == 'marker':
            marker_labels = list(GraphStyle.pyqtgraphMarkers.keys())
            marker_keys = list(GraphStyle.pyqtgraphMarkers.values())
            marker = self[key]
            if marker is None:
                marker = 'None'
            elif marker in marker_keys:
                marker = marker_labels[marker_keys.index(marker)]
            widget.setCurrentIndex(marker_labels.index(marker))

    def updateFromWidget(self, key: str, widget: QWidget) -> None:
        key = self.getKey(key)
        if key in ['color', 'markeredgecolor', 'markerfacecolor']:
            self[key] = widget.color()
        elif key in ['linestyle', 'markeredgestyle']:
            self[key] = GraphStyle.lineStyles[widget.currentIndex()]
        elif key in ['linewidth', 'markersize', 'markeredgewidth']:
            self[key] = widget.value()
        elif key == 'marker':
            self[key] = widget.currentText()


class GraphStylePanel(QWidget):

    def __init__(self, styles: list[str] = None, *args, **kwargs):
        QWidget.__init__(self, *args, **kwargs)

        if styles is None:
            styles = ['color', 'linestyle', 'linewidth', 'marker', 'markersize', 'markeredgestyle', 'markeredgewidth', 'markeredgecolor', 'markerfacecolor']
        styles = [GraphStyle.getKey(key) for key in styles]

        self._widgets: dict[str, QWidget] = {key: GraphStyle().createWidget(key) for key in styles}

        vbox = QVBoxLayout(self)
        vbox.setContentsMargins(10, 10, 10, 10)
        vbox.setSpacing(10)

        sections = [
            [
                'Line',
                ['color', 'linestyle', 'linewidth'],
                ['Color', 'Style', 'Width']
            ],
            [
                'Marker',
                ['marker', 'markersize', 'markeredgestyle', 'markeredgewidth', 'markeredgecolor', 'markerfacecolor'],
                ['Marker', 'Size', 'Edge Style', 'Edge Width', 'Edge Color', 'Face Color']
            ],
        ]
        for section in sections:
            title, keys, labels = section
            if set(keys).intersection(styles):
                sectionWidget = CollapsibleSection(title=title)
                form = QFormLayout()
                form.setContentsMargins(0, 0, 0, 0)
                form.setSpacing(6)
                form.setHorizontalSpacing(10)
                for key, label in zip(keys, labels):
                    if key in styles:
                        form.addRow(label, self._widgets[key])
                sectionWidget.setContentLayout(form)
                vbox.addWidget(sectionWidget)
        
        # expand 1st section
        if vbox.count() > 0:
            vbox.itemAt(0).widget().expand()
        
        vbox.addStretch()
    
    def graphStyle(self) -> GraphStyle:
        graphStyle = GraphStyle()
        for key, widget in self._widgets.items():
            graphStyle.updateFromWidget(key, widget)
        return graphStyle
    
    def setGraphStyle(self, graphStyle: GraphStyle):
        for key, widget in self._widgets.items():
            graphStyle.updateWidget(key, widget)


def editGraphStyle(graphStyle: GraphStyle, styles: list[str] = None, parent: QWidget = None, title: str = None) -> GraphStyle | None:
    panel = GraphStylePanel(styles)
    panel.layout().setContentsMargins(0, 0, 0, 0)
    panel.setGraphStyle(graphStyle)

    dlg = QDialog(parent)
    vbox = QVBoxLayout(dlg)
    vbox.addWidget(panel)

    btns = QDialogButtonBox()
    btns.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
    btns.accepted.connect(dlg.accept)
    btns.rejected.connect(dlg.reject)
    vbox.addWidget(btns)
    vbox.addStretch()

    if title is not None:
        dlg.setWindowTitle(title)
    dlg.setWindowModality(Qt.ApplicationModal)
    if dlg.exec() == QDialog.Accepted:
        return panel.graphStyle()


def test_live():
    app = QApplication()
    ui = GraphStylePanel()
    ui.show()
    # QTimer.singleShot(1000, lambda: print(editGraphStyle(GraphStyle())))
    app.exec()


if __name__ == '__main__':
    test_live()



""" PySide/PyQt 2-D plot.
"""

# from qtpy.QtCore import *
# from qtpy.QtGui import *
# from qtpy.QtWidgets import *
# from qtpy.QtCharts import *
# import numpy as np
# import platform

# # See https://github.com/PlotPyStack/PythonQwt/blob/master/qwt/plot_curve.py#L63
# import os
# try:
#     QT_API = os.environ["QT_API"].lower()
# except:
#     from qtpy import API_NAME
#     QT_API = API_NAME.lower()
# if QT_API == "pyside6":
#     import shiboken6 as shiboken
#     import ctypes


# class Plot(QChartView):
#     """ PySide/PyQt 2-D plot.
#     """
    
#     def __init__(self, *args, **kwargs):
#         QChartView.__init__(self, *args, **kwargs)

#         self._chart: QChart = QChart()
#         # self._chart.createDefaultAxes()
#         self._xaxis: QValueAxis = QValueAxis()
#         self._yaxis: QValueAxis = QValueAxis()
#         self._chart.addAxis(self._xaxis, Qt.AlignmentFlag.AlignBottom)
#         self._chart.addAxis(self._yaxis, Qt.AlignmentFlag.AlignLeft)
#         self.setChart(self._chart)

#         if platform.system() == 'Darwin':
#             # Fix error message due to touch events on MacOS trackpad.
#             # !!! Warning: This may break touch events on a touch screen or mobile device.
#             # See https://bugreports.qt.io/browse/QTBUG-103935
#             self.viewport().setAttribute(Qt.WidgetAttribute.WA_AcceptTouchEvents, False)
    

# def array2d_to_qpolygonf(xdata, ydata) -> QPolygonF:
#     """
#     See https://github.com/PlotPyStack/PythonQwt/blob/master/qwt/plot_curve.py#L63

#     Utility function to convert two 1D-NumPy arrays representing curve data
#     (X-axis, Y-axis data) into a single polyline (QtGui.PolygonF object).
#     This feature is compatible with PyQt5 and PySide6 (requires QtPy).

#     License/copyright: MIT License © Pierre Raybaut 2020-2021.

#     :param numpy.ndarray xdata: 1D-NumPy array
#     :param numpy.ndarray ydata: 1D-NumPy array
#     :return: Polyline
#     :rtype: QtGui.QPolygonF
#     """
#     if not (xdata.size == ydata.size == xdata.shape[0] == ydata.shape[0]):
#         raise ValueError("Arguments must be 1D NumPy arrays with same size")
#     size = xdata.size
#     if QT_API.startswith("pyside"):  # PySide (obviously...)
#         polyline = QPolygonF()
#         polyline.resize(size)
#         address = shiboken.getCppPointer(polyline.data())[0]
#         buffer = (ctypes.c_double * 2 * size).from_address(address)
#     else:  # PyQt
#         if QT_API == "pyqt6":
#             polyline = QPolygonF([QPointF(0, 0)] * size)
#         else:
#             polyline = QPolygonF(size)
#         buffer = polyline.data()
#         buffer.setsize(16 * size)  # 16 bytes per point: 8 bytes per X,Y value (float64)
#     memory = np.frombuffer(buffer, np.float64)
#     memory[: (size - 1) * 2 + 1 : 2] = np.array(xdata, dtype=np.float64, copy=False)
#     memory[1 : (size - 1) * 2 + 2 : 2] = np.array(ydata, dtype=np.float64, copy=False)
#     return polyline


# def test_live():
#     import sys
#     import time

#     app = QApplication(sys.argv)
#     plot = Plot()

#     series = QLineSeries(name='test')
#     plot._chart.addSeries(series)
#     series.attachAxis(plot._xaxis)
#     series.attachAxis(plot._yaxis)
#     series.setUseOpenGL(True)

#     a = time.time()
#     n = int(1e7)
#     x = np.linspace(0, 10, n)
#     y = np.random.rand(n)
#     b = time.time()
#     poly = array2d_to_qpolygonf(x, y)
#     points = QPointFList()
#     points.reserve(n)
#     for i in range(n):
#         points.append(poly[i])
#     c = time.time()
#     series.replace(points)
#     plot.show()
#     d = time.time()
#     print('data creation (sec):', b - a)
#     print('data -> line series (sec):', c - b)
#     print('line plot (sec):', d - c)

#     app.exec()


# def test_live_pyqtgraph():
#     import sys
#     import time
#     from pyqtgraph_ext import PlotWidget, XYDataItem

#     app = QApplication(sys.argv)
#     plot = PlotWidget()

#     a = time.time()
#     n = int(1e7)
#     x = np.linspace(0, 10, n)
#     y = np.random.rand(n)
#     b = time.time()
#     data = XYDataItem(x=x, y=y)
#     plot.addItem(data)
#     plot.show()
#     c = time.time()
#     print('data creation (sec):', b - a)
#     print('line plot (sec):', c - b)

#     app.exec()


# def test_live_vispy():
#     import sys
#     import time
#     from vispy import plot as vp

#     app = QApplication(sys.argv)
#     win = QMainWindow()
#     fig = vp.Fig()
#     win.setCentralWidget(fig.native)

#     a = time.time()
#     n = int(1e7)
#     x = np.linspace(0, 10, n)
#     y = np.random.rand(n)
#     b = time.time()
#     ax = fig[0,0]
#     ax.plot((x, y))
#     win.show()
#     c = time.time()
#     print('data creation (sec):', b - a)
#     print('line plot (sec):', c - b)

#     app.exec()


# if __name__ == '__main__':
#     # test_live()
#     # test_live_pyqtgraph()
#     test_live_vispy()




""" Base class for a tree model that uses AbstractTreeItem for its data interface.

Supports drag-and-drop within and between models in the same program/process.
"""

from __future__ import annotations
from warnings import warn
from typing import Any
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from pyqt_ext.tree import AbstractTreeItem


class AbstractTreeModel(QAbstractItemModel):
    """ Base class for a tree model that uses AbstractTreeItem for its data interface.

    Constructor takes either the tree data itself or a root item wrapping the tree data.

    This class can work as is, but in general you will derive from it and reimplement `setupItemTree`, `treeData`, `data` and `setData`. Optionally also reimplement `columnCount`, `flags`, etc. to suit your data.

    If needed, you can overwrite MIME_TYPE in a derived class to a custom MIME type for drag-and-drop operations.
    """

    MIME_TYPE = 'application/x-AbstractTreeModel'

    def __init__(self, data: AbstractTreeItem | Any = None, **kwargs):
        QAbstractItemModel.__init__(self, **kwargs)

        # tree data interface
        self._rootItem: AbstractTreeItem = None
        self.setTreeData(data)

        # headers
        self._rowLabels: list = []
        self._columnLabels: list = []

        # drag-and-drop
        self._supportedDropActions: Qt.DropActions = Qt.DropAction.MoveAction | Qt.DropAction.CopyAction
        self._mimeTypes: list[str] = [self.MIME_TYPE]
    
    # !! Must reimpliment.
    def setupItemTree(self, data: Any) -> AbstractTreeItem:
        """ Reimpliment in derived class to appropriately build an item tree from data.
        """
        if data is None:
            return
        if isinstance(data, AbstractTreeItem):
            return data
        # you must reimpliment this method to handle the input tree data...
        raise NotImplementedError
    
    # !! Must reimpliment.
    def treeData(self) -> Any:
        """ Reimpliment in derived class to appropriately return the underlying tree data.
        """
        raise NotImplementedError
    
    def setTreeData(self, data: Any) -> None:
        """ Rebuild the item tree for the input tree data.
        """
        if (data is None) or isinstance(data, AbstractTreeItem):
            self.setRootItem(data)
            return
        
        self.beginResetModel()
        self._rootItem = self.setupItemTree(data)
        self.endResetModel()
    
    def rootItem(self) -> AbstractTreeItem:
        return self._rootItem
    
    def setRootItem(self, root: AbstractTreeItem) -> None:
        self.beginResetModel()
        self._rootItem = root
        self.endResetModel()
    
    def reset(self) -> None:
        """ Reset the model.
        """
        try:
            self.setTreeData(self.treeData())
        except:
            self.beginResetModel()
            self.endResetModel()
    
    def rowCount(self, parent_index: QModelIndex = QModelIndex()) -> int:
        """ Uses `AbstractTreeItem.children` to get the number of children of parent.
        """
        if parent_index.column() > 0:
            # ?? Only first column index has children. Not sure this matters.
            return 0
        if not parent_index.isValid():
            parent_item: AbstractTreeItem = self.rootItem()
        else:
            parent_item: AbstractTreeItem = self.itemFromIndex(parent_index)
        if parent_item is None:
            return 0
        return len(parent_item.children)

    def columnCount(self, parent_index: QModelIndex = QModelIndex()) -> int:
        """ Defaults to a single column tree.

        Reimplement in a derived class if you need more than one column.
        """
        return 1

    def itemFromIndex(self, index: QModelIndex = QModelIndex()) -> AbstractTreeItem | None:
        """ Get the item associated with index.
        """
        if not index.isValid():
            return self.rootItem()
        item: AbstractTreeItem = index.internalPointer()
        return item
    
    def indexFromItem(self, item: AbstractTreeItem, column: int = 0) -> QModelIndex:
        """ Get the index associated with item.

        Each item is associated with a row, so the column is either assumed to be 0 or needs to be given.
        """
        if (item is self.rootItem()) or (item.parent() is None):
            # item is the root item
            return QModelIndex()
        row: int = item.siblingIndex()
        return self.createIndex(row, column, item)

    def parent(self, index: QModelIndex = QModelIndex()) -> QModelIndex:
        """ Uses item tree to get the parent index.
        """
        if not index.isValid():
            return QModelIndex()
        item: AbstractTreeItem = self.itemFromIndex(index)
        parent_item: AbstractTreeItem = item.parent()
        if parent_item is None or parent_item is self.rootItem():
            return QModelIndex()
        return self.indexFromItem(parent_item)

    def index(self, row: int, column: int, parent_index: QModelIndex = QModelIndex()) -> QModelIndex:
        """ Return an index associated with the appropriate AbstractTreeItem.
        """
        if not self.hasIndex(row, column, parent_index):
            return QModelIndex()
        if parent_index.isValid() and parent_index.column() != 0:
            return QModelIndex()
        try:
            parent_item: AbstractTreeItem = self.itemFromIndex(parent_index)
            item: AbstractTreeItem = parent_item.children[row]
            return self.createIndex(row, column, item)
        except IndexError:
            return QModelIndex()

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        """ Default item flags.
        
        Supports drag-and-drop if it is enabled in `supportedDropActions`.
        Reimplement in a derived class if you need custom flags.
        """
        if not index.isValid():
            # root item
            if self.supportedDropActions() != Qt.DropAction.IgnoreAction:
                # allow drops on the root item (i.e., this allows drops on the viewport away from other items)
                return Qt.ItemFlag.ItemIsDropEnabled
            return Qt.ItemFlag.NoItemFlags
        flags = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEditable
        if self.supportedDropActions() != Qt.DropAction.IgnoreAction:
            flags |= Qt.ItemFlag.ItemIsDragEnabled | Qt.ItemFlag.ItemIsDropEnabled
        return flags

    # !! Probably need to reimpliment.
    def data(self, index: QModelIndex, role: int):
        """ Reimpliment in a derived class to get the data at index.

        Default implementation returns the item name for the first column.
        """
        if not index.isValid():
            return
        item: AbstractTreeItem = self.itemFromIndex(index)
        if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
            if index.column() == 0:
                return item.name()

    # !! Probably need to reimpliment.
    def setData(self, index: QModelIndex, value, role: int) -> bool:
        """ Reimpliment in a derived class to set the data at index.

        Default implementation sets the item name for the first column.
        """
        if not index.isValid():
            return False
        item: AbstractTreeItem = self.itemFromIndex(index)
        if role == Qt.ItemDataRole.EditRole:
            if index.column() == 0:
                item.setName(value)
                self.dataChanged.emit(index, index)  # ?? needed?
                return True
        return False

    def headerData(self, section: int, orientation: Qt.Orientation, role: int):
        """ Get data from `rowLabels` or `columnLabels`.
        """
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                labels = self.columnLabels()
            elif orientation == Qt.Orientation.Vertical:
                labels = self.rowLabels()
            if section < len(labels):
                label = labels[section]
                if label is not None:
                    return label
            return section

    def setHeaderData(self, section: int, orientation: Qt.Orientation, value, role: int) -> bool:
        """ Set data in `rowLabels` or `columnLabels`.
        """
        if role == Qt.ItemDataRole.EditRole:
            if orientation == Qt.Orientation.Horizontal:
                labels = self.columnLabels()
            elif orientation == Qt.Orientation.Vertical:
                labels = self.rowLabels()
            if section < len(labels):
                labels[section] = value
            else:
                labels += [None] * (section - len(labels)) + [value]
            if orientation == Qt.Orientation.Horizontal:
                self.setColumnLabels(labels)
            elif orientation == Qt.Orientation.Vertical:
                self.setRowLabels(labels)
            self.headerDataChanged.emit(orientation, section, section)
            return True
        return False
    
    def rowLabels(self) -> list:
        return self._rowLabels
    
    def setRowLabels(self, labels: list) -> None:
        old_labels = self._rowLabels
        n_overlap = min(len(labels), len(old_labels))
        first_change = 0
        while (first_change < n_overlap) and (labels[first_change] == old_labels[first_change]):
            first_change += 1
        last_change = max(len(labels), len(old_labels)) - 1
        while (last_change < n_overlap) and (labels[last_change] == old_labels[last_change]):
            last_change -= 1
        self._rowLabels = labels
        if first_change <= last_change: 
            self.headerDataChanged.emit(Qt.Orientation.Vertical, first_change, last_change)
    
    def columnLabels(self) -> list | None:
        return self._columnLabels
    
    def setColumnLabels(self, labels: list | None) -> None:
        old_labels = self._columnLabels
        n_overlap = min(len(labels), len(old_labels))
        first_change = 0
        while (first_change < n_overlap) and (labels[first_change] == old_labels[first_change]):
            first_change += 1
        last_change = max(len(labels), len(old_labels)) - 1
        while (last_change < n_overlap) and (labels[last_change] == old_labels[last_change]):
            last_change -= 1
        self._columnLabels = labels
        if first_change <= last_change: 
            self.headerDataChanged.emit(Qt.Orientation.Horizontal, first_change, last_change)
    
    def maxDepth(self):
        root: AbstractTreeItem = self.rootItem()
        if root is None:
            return 0
        return root.maxDepthBelow()

    def removeRows(self, row: int, count: int, parent_index: QModelIndex = QModelIndex()) -> bool:
        """ Calls `AbstractTreeItem.removeChild` to remove rows.
        
        Reimpliment if you need more efficient removal of multiple rows than the default one row at a time, or if you need custom logic to check the validity of the removals.
        Otherwise you probably do not need to touch this if you have appropriately implemented `AbstractTreeItem.removeChild`.
        """
        if count <= 0:
            return False
        n_rows: int = self.rowCount(parent_index)
        if (row < 0) or (row + count > n_rows):
            return False
        parent_item: AbstractTreeItem = self.itemFromIndex(parent_index)
        self.beginRemoveRows(parent_index, row, row + count - 1)
        # remove items in reverse order to preserve the sibling index of subsequent items to remove
        for i in reversed(list(range(row, row + count))):
            item: AbstractTreeItem = parent_item.children[i]
            parent_item.removeChild(item)
        self.endRemoveRows()
        return True
    
    def insertRows(self, row: int, count: int, parent_index: QModelIndex = QModelIndex()) -> bool:
        """ Calls `AbstractTreeItem.insertChild` to insert rows.
        
        This only inserts `AbstractTreeItem`s. To insert derived classes, see `insertItems`.
        """
        if count <= 0:
            return False
        n_rows: int = self.rowCount(parent_index)
        if (row < 0) or (row > n_rows):
            return False
        parent_item: AbstractTreeItem = self.itemFromIndex(parent_index)
        self.beginInsertRows(parent_index, row, row + count - 1)
        for i in range(row, row + count):
            parent_item.insertChild(i, AbstractTreeItem())
        self.endInsertRows()
        return True
    
    def moveRows(self, src_parent_index: QModelIndex, src_row: int, count: int, dst_parent_index: QModelIndex, dst_row: int = -1) -> bool:
        """ Calls `AbstractTreeItem.insertChild` to move items (e.g., rows) within the tree.
        
        Reimpliment if you need custom logic to check the validity of the move.
        Otherwise you probably do not need to touch this if you have appropriately implemented `AbstractTreeItem.insertChild`.
        """
        n_src_rows: int = self.rowCount(src_parent_index)
        n_dst_rows: int = self.rowCount(dst_parent_index)
        if (src_row < 0) or (src_row + count > n_src_rows):
            return False
        if dst_row == -1:
            # append
            dst_row = n_dst_rows
        elif not (0 <= dst_row <= n_dst_rows):
            return False
        if (src_parent_index == dst_parent_index) and (0 <= dst_row - src_row <= count):
            # no change, dst is within src items
            return False
        
        src_parent_item: AbstractTreeItem = self.itemFromIndex(src_parent_index)
        dst_parent_item: AbstractTreeItem = self.itemFromIndex(dst_parent_index)

        src_items: list[AbstractTreeItem] = src_parent_item.children[src_row: src_row + count]

        for src_item in src_items:
            if dst_parent_item.hasAncestor(src_item):
                warn('Cannot move an item to one of its descendants.')
                return False

        self.beginMoveRows(src_parent_index, src_row, src_row + count - 1, dst_parent_index, dst_row)
        for i, src_item in enumerate(src_items):
            dst_parent_item.insertChild(dst_row + i, src_item)
            if src_parent_item is dst_parent_item:
                if src_row < dst_row:
                    dst_row -= 1
        self.endMoveRows()
        return True
    
    def removeItems(self, items: list[AbstractTreeItem]) -> bool:
        """ Calls `removeRows` to remove items from the model.

        !! The boolean return value only indicates the success for the removal of the last contiguous block of items, and thus only makes complete sense if removing a single contiguous block of items.
        """
        success: bool = False
        item_groups: list[list[AbstractTreeItem]] = self.groupItems(items)
        for item_group in item_groups:
            parent_item: AbstractTreeItem = item_group[0].parent()
            parent_index: QModelIndex = self.indexFromItem(parent_item)
            row: int = item_group[0].siblingIndex()
            count: int = len(item_group)
            success = self.removeRows(row, count, parent_index)
        # !! return value only makes complete sense if removing a single contiguous block of items
        return success
    
    def removeItem(self, item: AbstractTreeItem) -> bool:
        return self.removeItems([item])
    
    def insertItems(self, row: int, items: list[AbstractTreeItem], parent_item: AbstractTreeItem) -> bool:
        """ Calls `AbstractTreeItem.insertChild` to insert items (i.e., rows).
        
        Rempliment if you need custom logic to ensure validity of insertions.
        Otherwise you probably do not need to touch this if you have appropriately implemented `AbstractTreeItem.insertChild`.
        """
        if not items:
            return False
        n_rows: int = len(parent_item.children)
        if row == -1:
            row = n_rows  # append
        elif (row < 0) or (row > n_rows):
            return False
        parent_index: QModelIndex = self.indexFromItem(parent_item)
        count: int = len(items)
        self.beginInsertRows(parent_index, row, row + count - 1)
        for i, item in enumerate(items):
            parent_item.insertChild(row + i, item)
        self.endInsertRows()
        return True
    
    def insertItem(self, row: int, item: AbstractTreeItem, parent_item: AbstractTreeItem) -> bool:
        return self.insertItems(row, [item], parent_item)
    
    def appendItems(self, items: list[AbstractTreeItem], parent_item: AbstractTreeItem) -> bool:
        """ Calls `insertItems` to append child items.
        """
        row = len(parent_item.children)
        return self.insertItems(row, items, parent_item)
    
    def appendItem(self, item: AbstractTreeItem, parent_item: AbstractTreeItem) -> bool:
        return self.appendItems([item], parent_item)
    
    def moveItems(self, items: list[AbstractTreeItem], dst_parent_item: AbstractTreeItem, dst_row: int = -1) -> bool:
        """ Calls `moveRows` to move items within the model.

        !! The boolean return value only indicates the success for moving the last contiguous block of items, and thus only makes complete sense if moving a single contiguous block of items.
        """
        success: bool = False
        dst_parent_index: QModelIndex = self.indexFromItem(dst_parent_item)
        item_groups: list[list[AbstractTreeItem]] = self.groupItems(items)
        for item_group in item_groups:
            src_parent_item: AbstractTreeItem = item_group[0].parent()
            src_parent_index: QModelIndex = self.indexFromItem(src_parent_item)
            src_row: int = item_group[0].siblingIndex()
            count: int = len(item_group)
            success = self.moveRows(src_parent_index, src_row, count, dst_parent_index, dst_row)
        # !! return value only makes complete sense if moving a single contiguous block of items
        return success
    
    def moveItem(self, item: AbstractTreeItem, dst_parent_item: AbstractTreeItem, dst_row: int = -1) -> bool:
        return self.moveItems([item], dst_parent_item, dst_row)
    
    def transferItems(self, src_model: AbstractTreeModel, src_items: list[AbstractTreeItem], dst_model: AbstractTreeModel, dst_parent_item: AbstractTreeItem, dst_row: int = -1) -> bool:
        """ Moves items between models.

        !! The boolean return value only indicates the success for moving the last contiguous block of items, and thus only makes complete sense if moving a single contiguous block of items.
        """
        if src_model is dst_model:
            return dst_model.moveItems(src_items, dst_parent_item, dst_row)

        success: bool = False
        dst_parent_index: QModelIndex = dst_model.indexFromItem(dst_parent_item)
        src_item_groups: list[list[AbstractTreeItem]] = src_model.groupItems(src_items)
        for item_group in src_item_groups:
            # remove items from src_model
            src_parent_item: AbstractTreeItem = item_group[0].parent()
            src_parent_index: QModelIndex = src_model.indexFromItem(src_parent_item)
            src_row: int = item_group[0].siblingIndex()
            count: int = len(item_group)
            success = src_model.removeRows(src_row, count, src_parent_index)
            if not success:
                continue
            # insert itmes in dst_model
            success = dst_model.insertItems(dst_row, item_group, dst_parent_item)
        # !! return value only makes complete sense if moving a single contiguous block of items
        return success
    
    def transferItem(self, src_model: AbstractTreeModel, src_item: AbstractTreeItem, dst_model: AbstractTreeModel, dst_parent_item: AbstractTreeItem, dst_row: int = -1) -> bool:
        self.transferItems(src_model, [src_item], dst_model, dst_parent_item, dst_row)
    
    def groupItems(self, items: list[AbstractTreeItem]) -> list[list[AbstractTreeItem]]:
        """ Group items by their parent item and contiguous row blocks.
        
        This is used to prepare for moving or copying items in the tree.
        Note: items should never contain the root item.
        """
        root: AbstractTreeItem = self.rootItem()
        if root in items:
            items.remove(root)
        group_parents: list[AbstractTreeItem] = []
        item_groups: list[list[AbstractTreeItem]] = []
        # group items in reverse depth-first order so that moving/removing them in order does not change the location of the remaining items
        for item in root.reverseDepthFirst():
            if item not in items:
                continue
            added = False
            for group_parent, item_group in zip(group_parents, item_groups):
                if group_parent is item.parent():
                    row: int = item.siblingIndex()
                    if row == item_group[0].siblingIndex() - 1:
                        item_group.insert(0, item)
                        added = True
                        break
            if not added:
                # start a new group
                group_parents.append(item.parent())
                item_groups.append([item])
        return item_groups
    
    def supportedDropActions(self) -> Qt.DropActions:
        return self._supportedDropActions
    
    def setSupportedDropActions(self, actions: Qt.DropActions) -> None:
        self._supportedDropActions = actions
    
    def mimeTypes(self) -> list[str]:
        """ Return the MIME types supported by this view for drag-and-drop operations.
        """
        return self._mimeTypes
    
    def setMimeTypes(self, mime_types :list[str]) -> None:
        self._mimeTypes = mime_types

    def mimeData(self, indexes: list[QModelIndex]) -> AbstractTreeMimeData | None:
        if not indexes:
            return None
        if self.rootItem() is None:
            return None
        items: list[AbstractTreeItem] = [self.itemFromIndex(index) for index in indexes if index.isValid()]
        if not items:
            return None
        return AbstractTreeMimeData(self, items, self.MIME_TYPE)

    def dropMimeData(self, data: AbstractTreeMimeData, action: Qt.DropAction, row: int, column: int, parent_index: QModelIndex) -> bool:
        if not isinstance(data, AbstractTreeMimeData):
            return False
        if not data.hasFormat(self.MIME_TYPE):
            return False
        
        src_model: AbstractTreeModel = data.model
        src_items: list[AbstractTreeItem] = data.items
        if not src_model or not src_items:
            return False

        # move src_items to the destination (row-th child of parent_index)
        dst_model: AbstractTreeModel = self
        if dst_model.rootItem() is None:
            return False
        dst_parent_item: AbstractTreeItem = dst_model.itemFromIndex(parent_index)
        
        self.transferItems(src_model, src_items, dst_model, dst_parent_item, row)

        # !? If we return True, the model will attempt to remove rows.
        # As we already completely handled the move, this will corrupt our model, so return False.
        return False
    
    def popupWarningDialog(self, text: str, title: str = 'Warning') -> None:
        """ Popup warning dialog at the currently focused widget.
        """
        focused_widget: QWidget = QApplication.focusWidget()
        QMessageBox.warning(focused_widget, title, text)
    
    """
    Everything below here is for path-based manipulation of the tree.
    !! This typically only makes sense if all sibling items have unique names.
       If sibling items should be able to have the same name, you probably should not use the functions below.
    """
    
    def pathFromItem(self, item: AbstractTreeItem) -> str:
        """ Path from the chain of item names up to root.
        """
        return item.path()
    
    def itemFromPath(self, path: str) -> AbstractTreeItem:
        """ Find the item associated with path starting from root.
        """
        root: AbstractTreeItem = self.rootItem()
        return root[path]

    def pathFromIndex(self, index: QModelIndex = QModelIndex()) -> str:
        """ Get the path associated with index.
        """
        if not index.isValid():
            # root item
            return '/'
        item: AbstractTreeItem = self.itemFromIndex(index)
        return self.pathFromItem(item)
    
    def indexFromPath(self, path: str) -> QModelIndex:
        """ Get the index associated with path.
        """
        item: AbstractTreeItem = self.itemFromPath(path)
        return self.indexFromItem(item)

    def isValidPath(self, path: str) -> bool:
        try:
            self.itemFromPath(path)
            return True
        except:
            return False
    
    def parentPath(self, path: str) -> str | None:
        """ Get the parent path for a given path.
        """
        if path == '/':
            return None
        parts = path.split('/')
        parent_path = '/'.join(parts[:-1])
        return parent_path or '/'
    
    # def groupPaths(self, paths: list[str]) -> list[list[str]]:
    #     """ Group paths by their parent item and contiguous row blocks.
        
    #     This is used to prepare for moving or copying paths in the tree.
    #     Note: paths should never contain the root path '/'.
    #     """
    #     if '/' in paths:
    #         paths.remove('/')
    #     paths = [path for path in paths if self.isValidPath(path)]
    #     items: list[AbstractTreeItem] = [self.itemFromPath(path) for path in paths]
    #     item_groups: list[list[AbstractTreeItem]] = self.groupItems(items)
    #     path_groups: list[list[str]] = [[self.pathFromItem(item) for item in item_group] for item_group in item_groups]
    #     return path_groups
    #     # # groups[parent path][[block row indices], ...]
    #     # groups: dict[str, list[list[int]]] = {}
    #     # for path in paths:
    #     #     item: AbstractTreeItem = self.itemFromPath(path)
    #     #     parent_item: AbstractTreeItem = item.parent()  # will be valid
    #     #     parent_path: str = parent_item.path()
    #     #     if parent_path not in groups:
    #     #         groups[parent_path] = []
    #     #     row: int = item.siblingIndex()
    #     #     added = False
    #     #     rows_group: list[int]
    #     #     for rows_group in groups[parent_path]:
    #     #         if row == rows_group[0] - 1:
    #     #             rows_group.insert(0, row)
    #     #             added = True
    #     #             break
    #     #         elif row == rows_group[-1] + 1:
    #     #             rows_group.append(row)
    #     #             added = True
    #     #             break
    #     #     if not added:
    #     #         # start a new group of rows
    #     #         groups[parent_path].append([row])

    #     # # order groups by parent path from leaves to root so that moving/removing them in order does not change the paths of the remaining groups
    #     # parent_paths: list[str] = list(groups.keys())
    #     # parent_paths.sort(key=lambda p: p.count('/'), reverse=True)  # sort by depth (more slashes means deeper in the tree)
    #     # groups = {parent_path: groups[parent_path] for parent_path in parent_paths}

    #     # # order row blocks from last block to first block so that removing them in order does not change the row indices of the remaining blocks
    #     # for parent_path, row_blocks in groups.items():
    #     #     row_blocks.sort(key=lambda block: block[0], reverse=True)
        
    #     # return groups
    
    # def removePaths(self, paths: list[str]) -> None:
    #     """ Remove items by their path.
    #     """
    #     # group paths for removal block by contiguous row block
    #     path_groups: list[list[str]] = self.groupPaths(paths)
    #     for path_groups in path_groups:
    #         for row_block in row_blocks:
    #             row: int = row_block[0]
    #             count: int = len(row_block)
    #             parent_index: QModelIndex = self.indexFromPath(parent_path)
    #             self.removeRows(row, count, parent_index)
    
    # def movePaths(self, src_paths: list[str], dst_parent_path: str, dst_row: int = -1) -> None:
    #     """ Move items within tree by their path.
    #     """
    #     dst_parent_index: QModelIndex = self.indexFromPath(dst_parent_path)
    #     dst_num_rows: int = self.rowCount(dst_parent_index)
    #     if (dst_row < 0) or (dst_row > dst_num_rows):
    #         # append rows
    #         dst_row = dst_num_rows
        
    #     # group paths for moving block by contiguous row block
    #     src_path_groups: dict[str, list[list[int]]] = self.groupPaths(src_paths)
    #     for src_parent_path, row_blocks in src_path_groups.items():
    #         for row_block in row_blocks:
    #             row: int = row_block[0]
    #             count: int = len(row_block)
    #             src_parent_index: QModelIndex = self.indexFromPath(src_parent_path)
    #             self.moveRows(src_parent_index, row, count, dst_parent_index, dst_row)
    
    # def transferPaths(self, src_model: AbstractTreeModel, src_paths: list[str], dst_model: AbstractTreeModel, dst_parent_path: str, dst_row: int = -1) -> None:
    #     """ Move items by path between two different `AbstractTreeModel`s.
    #     """
    #     if src_model is dst_model:
    #         # move within the same tree model
    #         src_model.movePaths(src_paths, dst_parent_path, dst_row)
    #         return
        
    #     # TODO... implement transfer between different tree models
    #     raise NotImplementedError


class AbstractTreeMimeData(QMimeData):
    """ Custom MIME data class for drag-and-drop with `AbstractTreeModel`s.

    This class allows storing a reference to an `AbstractTreeModel` and some of its `AbstractTreeItem`s (i.e., the items being dragged during drag-and-drop) in the MIME data.
    This allows simple transfer of tree items within and between `AbstractTreeModel`s in the same program/process.

    Note:
    This approach probably won't work if you need to pass items between `AbstractTreeModel`s in separate programs/processes.
    If you really need to do this, you need to somehow serialize the dragged items (maybe with pickle), pass the serialized bytes in the drag MIME data, then deserialize back to the items on drop.
    """

    def __init__(self, model: AbstractTreeModel, items: list[AbstractTreeItem], mimeType: str):
        QMimeData.__init__(self)

        # these define the tree items being dragged and their MIME type
        self.model: AbstractTreeModel = model
        self.items: list[AbstractTreeItem] = items
        self.mimeType = mimeType

        # The actual value of the data here is not important, as we won't use it.
        # Instead, we will use the above attributes to handle drag-and-drop.
        # This is just to ensure that the MIME data has a valid format.
        self.setData(self.mimeType, self.mimeType.encode('utf-8'))
    
    def hasFormat(self, mimeType: str) -> bool:
        """ Check if the MIME data has the specified format.
        
        Overrides the default method to check for self.mimeType.
        """
        return mimeType == self.mimeType or super().hasFormat(mimeType)


def test_model():
    root = AbstractTreeItem()
    AbstractTreeItem(parent=root)
    root.appendChild(AbstractTreeItem(name='child2'))
    root.insertChild(1, AbstractTreeItem(name='child3'))
    root.children[1].appendChild(AbstractTreeItem())
    root.children[1].appendChild(AbstractTreeItem())
    root.children[1].appendChild(AbstractTreeItem())
    grandchild2 = AbstractTreeItem(name='grandchild2')
    grandchild2.setParent(root['child2'])
    AbstractTreeItem(name='greatgrandchild', parent=root['/child2/grandchild2'])

    print('\nInitial model...')
    model = AbstractTreeModel(root)
    print(model.rootItem())

    print('\nRemove grandchild2...')
    model.removeItem(grandchild2)
    print(model.rootItem())

    print('\nInsert grandchild2...')
    model.insertItems(0, [grandchild2], root['child2'])
    print(model.rootItem())


if __name__ == '__main__':
    test_model()


""" PySide/PyQt multi-value spin box.
"""

from __future__ import annotations
from qtpy.QtCore import Qt, Signal, Slot
from qtpy.QtGui import QValidator
from qtpy.QtWidgets import QAbstractSpinBox, QSizePolicy, QApplication
import re
import numpy as np


class MultiValueSpinBox(QAbstractSpinBox):
    """ Spinbox allowing multiple values or value ranges.

    - Specify allowed values as a list or numpy array.
    - Select multiple space or comma-separated values or value ranges.
    - Value ranges can be specified as first-last or start:stop[:step].
    """

    indicesChanged = Signal()

    def __init__(self, *args, **kwargs):
        QAbstractSpinBox.__init__(self, *args, **kwargs)

        # whether or not to show all individual values or values ranges when possible
        self._display_value_ranges_when_possible = True

        # suffix
        self._suffix = ''

        # possible values to select from
        self._indexed_values: np.ndarray = np.arange(100)

        # indices of selected values in self._indexed_values
        # i.e., selected values are self._indexed_values[self._indices]
        self._indices: np.ndarray[int] = np.array([0], dtype=int)

        # initialize with default values
        self.setIndices(self._indices)

        self.setSizePolicy(QSizePolicy.Expanding, self.sizePolicy().verticalPolicy())

        self.editingFinished.connect(self.onTextEdited)

        self.setToolTip('index/slice (+Shift: page up/down)')

        self.lineEdit().setPlaceholderText(self.textFromValues(self._indexed_values))
    
    def indices(self) -> np.ndarray[int]:
        mask = (self._indices >= 0) & (self._indices < len(self._indexed_values))
        return self._indices[mask]
    
    def setIndices(self, indices: list[int] | np.ndarray[int]):
        # print('set_indices:', indices, 'into', self._indexed_values)
        if not isinstance(indices, np.ndarray):
            indices = np.array(indices, dtype=int)
        elif not np.issubdtype(indices.dtype, np.integer):
            indices = indices.astype(int)
        mask = (indices >= 0) & (indices < len(self._indexed_values))
        self._indices = indices[mask]
        text = self.textFromValues(self.selectedValues())
        self.lineEdit().setText(text)
        self.indicesChanged.emit()
    
    def indexedValues(self) -> np.ndarray:
        return self._indexed_values
    
    def setIndexedValues(self, values: list | np.ndarray):
        if not isinstance(values, np.ndarray):
            dtype = type(values[0]) if len(values) > 0 else int
            values = np.array(values, dtype=dtype)
        self._indexed_values = values
        self.setIndices(self.indices())
        self.lineEdit().setPlaceholderText(self.textFromValues(self._indexed_values))
    
    def selectedValues(self) -> np.ndarray:
        return self._indexed_values[self.indices()]
    
    def setSelectedValues(self, values: list | np.ndarray):
        indices = self.indicesFromValues(values)
        self.setIndices(indices)
    
    def displayValueRangesWhenPossible(self) -> bool:
        return self._display_value_ranges_when_possible
    
    def setDisplayValueRangesWhenPossible(self, show_ranges: bool):
        self._display_value_ranges_when_possible = show_ranges
        # update text by resetting indices
        self.setIndices(self.indices())
    
    def suffix(self) -> str:
        return self._suffix
    
    def setSuffix(self, suffix: str):
        self._suffix = suffix
        # update text by resetting indices
        self.setIndices(self.indices())
    
    def indicesFromValues(self, values: list | np.ndarray, side='left', tol: float | None = None) -> np.ndarray[int]:
        if not isinstance(values, np.ndarray):
            values = np.array(values, dtype=self._indexed_values.dtype)
        if np.issubdtype(self._indexed_values.dtype, np.floating):
            indices = np.searchsorted(self._indexed_values, values, side=side)
            for i, index in enumerate(indices):
                if index > 0:
                    # check if previous value is closer
                    if np.abs(values[i] - self._indexed_values[index-1]) < np.abs(values[i] - self._indexed_values[index]):
                        indices[i] = index - 1
            if tol is not None:
                # only accept values within a tolerance, otherwise return all the closest indexed values
                mask = np.abs(values - self._indexed_values[indices]) <= tol
                indices = indices[mask]
        else:
            # exact matches only for non-floating point types
            if np.issubdtype(self._indexed_values.dtype, np.integer):
                values = np.intersect1d(self._indexed_values, values)
                indices = np.searchsorted(self._indexed_values, values)
            else:
                # do not assume values are ordered if not integer or floating point
                indices = []
                for value in values:
                    try:
                        i = np.where(self._indexed_values == value)[0][0]
                        indices.append(i)
                    except:
                        pass
                indices = np.array(indices, dtype=int)
        return indices
    
    def valuesFromText(self, text: str, validate: bool = False) -> list:
        if self.suffix():
            text = text.replace(self.suffix().strip(), '')
        text = text.strip()
        if text == '':
            return self._indexed_values
        if text == ':':
            return self._indexed_values
        fields = re.split(r'[,\s]+', text)
        values = []
        dtype = self._indexed_values.dtype.type
        for field in fields:
            try:
                field = field.strip()
                if field == '':
                    continue
                if field == ':':
                    return self._indexed_values
                if ':' in field:
                    # first:last inclusive
                    first, last = [dtype(arg) if len(arg.strip()) else None for arg in field.split(':')]
                    
                    if first is None:
                        start_index: int = 0
                    elif np.issubdtype(dtype, np.floating):
                        start_index: int = np.where(self._indexed_values >= first)[0][0]
                    else:
                        # exact match only
                        start_index: int = np.where(self._indexed_values == first)[0][0]
                    
                    if last is None:
                        stop_index: int = len(self._indexed_values)
                    elif np.issubdtype(dtype, np.floating):
                        stop_index: int = np.where(self._indexed_values <= last)[0][-1] + 1
                    else:
                        # exact match only
                        stop_index: int = np.where(self._indexed_values == last)[0][-1] + 1
                    
                    if stop_index > start_index:
                        values.extend(self._indexed_values[start_index:stop_index].tolist())
                else:
                    value = dtype(field)
                    index: int = self.indicesFromValues([value])[0]
                    closest_value = self._indexed_values[index]
                    values.append(closest_value)
            except:
                if validate:
                    raise ValueError(f'Invalid text: {field}')
                else:
                    continue
        values = np.array(values, dtype=dtype)
        return values
    
    def textFromValues(self, values: list | np.ndarray):
        indices = self.indicesFromValues(values)
        index_ranges = []
        for index in indices:
            if (not self._display_value_ranges_when_possible) or (len(index_ranges) == 0) or (index != index_ranges[-1][-1] + 1):
                index_ranges.append([index])
            else:
                index_ranges[-1].append(index)
        texts = []
        for index_range in index_ranges:
            if len(index_range) == 1:
                value = self._indexed_values[index_range[0]]
                if np.issubdtype(self._indexed_values.dtype, np.floating):
                    value_text = f'{value:.6f}'.rstrip('0').rstrip('.')
                else:
                    value_text = str(value)
                texts.append(value_text)
            else:
                first_value = self._indexed_values[index_range[0]]
                last_value = self._indexed_values[index_range[-1]]
                if np.issubdtype(self._indexed_values.dtype, np.floating):
                    first_value_text = f'{first_value:.6f}'.rstrip('0').rstrip('.')
                    last_value_text = f'{last_value:.6f}'.rstrip('0').rstrip('.')
                else:
                    first_value_text = str(first_value)
                    last_value_text = str(last_value)
                texts.append(first_value_text + ':' + last_value_text)
        text = ','.join(texts)
        if self.suffix():
            text += self.suffix()
        return text
    
    @Slot()
    def onTextEdited(self):
        text = self.lineEdit().text()
        try:
            values = self.valuesFromText(self.lineEdit().text(), validate=True)
            self.setSelectedValues(values)
        except:
            # do not overwrite text if invalid
            text = self.textFromValues(self.selectedValues())
            self.lineEdit().setText(text)
    
    def stepEnabled(self):
        indices = self.indices()
        min_index = 0
        max_index = len(self._indexed_values) - 1
        if len(indices) == 1:
            if indices[0] == min_index:
                return QAbstractSpinBox.StepEnabledFlag.StepUpEnabled
            if indices[0] == max_index:
                return QAbstractSpinBox.StepEnabledFlag.StepDownEnabled
        return QAbstractSpinBox.StepEnabledFlag.StepUpEnabled | QAbstractSpinBox.StepEnabledFlag.StepDownEnabled
    
    def stepBy(self, steps: int):
        indices = self.indices()
        min_index = 0
        max_index = len(self._indexed_values) - 1
        if len(indices) == 0:
            index = min_index if steps >= 0 else max_index
            indices = [index]
        elif len(indices) == 1:
            index = min(max(min_index, indices[0] + steps), max_index)
            indices = [index]
        else:
            modifiers = QApplication.keyboardModifiers()
            if modifiers & Qt.ShiftModifier:
                # page up/down
                num_indices = len(indices)
                if steps >= 0:
                    last = min(indices[-1] + num_indices, max_index)
                    first = max(min_index, last - num_indices + 1)
                else:
                    first = max(min_index, indices[0] - num_indices)
                    last = min(first + num_indices - 1, max_index)
                indices = list(range(first, last + 1))
            else:
                # step to end of current selected range 
                index = indices[0] if steps >= 0 else indices[-1]
                indices = [index]
        self.setIndices(indices)
    
    def validate(self, text, pos):
        # try:
        #     self.values_from_text(self.lineEdit().text(), validate=True)
        # except:
        #     return QValidator.Intermediate, text, pos
        return QValidator.Acceptable, text, pos

    
def test_live():
    from qtpy.QtWidgets import QApplication, QWidget, QVBoxLayout, QLabel, QCheckBox

    app = QApplication()

    def print_indices_and_values(spinbox: MultiValueSpinBox):
        print('indices:', spinbox.indices())
        print('selected_values:', spinbox.selectedValues())

    ui = QWidget()
    vbox = QVBoxLayout(ui)

    vbox.addWidget(QLabel('Try selecting multiple space- or comma-separated values.'))
    vbox.addWidget(QLabel('Try selecting value ranges as first:last.'))
    vbox.addWidget(QLabel('With multiple values selected, try stepping with Shift pressed.'))
    cbox = QCheckBox('Show value ranges when possible')
    cbox.setChecked(True)
    cbox.setToolTip('Change this and see how contiguous selected values are displayed.')
    vbox.addWidget(cbox)
    vbox.addStretch()

    spinboxes = []

    spinbox = MultiValueSpinBox()
    spinbox.setIndexedValues([0, 1, 2, 3, 4, 7, 8, 9])
    spinbox.indicesChanged.connect(lambda obj=spinbox: print_indices_and_values(obj))
    spinboxes.append(spinbox)
    vbox.addWidget(QLabel('0,1,2,3,4,7,8,9'))
    vbox.addWidget(spinbox)
    vbox.addStretch()

    spinbox = MultiValueSpinBox()
    spinbox.setIndexedValues([5,8,15,20])
    spinbox.setIndices([1,2])
    spinbox.indicesChanged.connect(lambda obj=spinbox: print_indices_and_values(obj))
    spinboxes.append(spinbox)
    vbox.addWidget(QLabel('5,8,15,20'))
    vbox.addWidget(spinbox)
    vbox.addStretch()

    spinbox = MultiValueSpinBox()
    spinbox.setIndexedValues(np.linspace(0,1,11))
    spinbox.indicesChanged.connect(lambda obj=spinbox: print_indices_and_values(obj))
    spinboxes.append(spinbox)
    vbox.addWidget(QLabel('0,.1,.2,.3,.4,.5,.6,.7,.8,.9,1'))
    vbox.addWidget(spinbox)
    vbox.addStretch()

    spinbox = MultiValueSpinBox()
    spinbox.setIndexedValues(['a','b','c','d','e','f'])
    spinbox.indicesChanged.connect(lambda obj=spinbox: print_indices_and_values(obj))
    spinboxes.append(spinbox)
    vbox.addWidget(QLabel('a,b,c,d,e,f'))
    vbox.addWidget(spinbox)
    vbox.addStretch()

    spinbox = MultiValueSpinBox()
    spinbox.setIndexedValues(['cat','mouse','dog','house','car','truck'])
    spinbox.indicesChanged.connect(lambda obj=spinbox: print_indices_and_values(obj))
    spinboxes.append(spinbox)
    vbox.addWidget(QLabel('cat, mouse, dog, house, car, truck'))
    vbox.addWidget(spinbox)
    vbox.addStretch()

    for spinbox in spinboxes:
        spinbox.setDisplayValueRangesWhenPossible(cbox.isChecked())
    
    def on_cbox_changed(checked):
        for spinbox in spinboxes:
            spinbox.setDisplayValueRangesWhenPossible(checked)
    
    cbox.stateChanged.connect(on_cbox_changed)

    ui.show()
    app.exec()


if __name__ == '__main__':
    test_live()



""" Tree model for a tree of key:value pairs that uses `KeyValueTreeItem`for its data interface.

e.g., Any level of nesting of dict and list objects.
Supports drag-and-drop within and between models in the same program/process.
"""

from __future__ import annotations
import numpy as np
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from pyqt_ext.tree import AbstractTreeModel, AbstractTreeMimeData, KeyValueTreeItem
import qtawesome as qta


class KeyValueTreeModel(AbstractTreeModel):

    MIME_TYPE = 'application/x-KeyValueTreeModel'
    
    def __init__(self, data: KeyValueTreeItem | dict | list = None, **kwargs):
        AbstractTreeModel.__init__(self, data, **kwargs)
        self.setColumnLabels(['Key', 'Value'])
    
    def setupItemTree(self, data: dict | list) -> KeyValueTreeItem:
        """ Build an item tree from data.
        """
        if data is None:
            return
        if isinstance(data, KeyValueTreeItem):
            return data
        # setup key:value item tree
        if not isinstance(data, dict) and not isinstance(data, list):
            raise ValueError('Root must be a dict or list.')
        return KeyValueTreeItem(None, data)
    
    def treeData(self) -> dict | list:
        """ The tree data.
        """
        root: KeyValueTreeItem = self.rootItem()
        root_map: dict | list =  root.value()
        return root_map
    
    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return 2

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        """ Default item flags.
        
        Supports drag-and-drop if it is enabled in `supportedDropActions`.
        """
        if not index.isValid():
            # root item
            if self.supportedDropActions() != Qt.DropAction.IgnoreAction:
                # allow drops on the root item (i.e., this allows drops on the viewport away from other items)
                return Qt.ItemFlag.ItemIsDropEnabled
            return Qt.ItemFlag.NoItemFlags
        flags = Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEditable
        if self.supportedDropActions() != Qt.DropAction.IgnoreAction:
            flags |= Qt.ItemFlag.ItemIsDragEnabled
            item: KeyValueTreeItem = self.itemFromIndex(index)
            value = item.value()
            if isinstance(value, dict) or isinstance(value, list):
                flags |= Qt.ItemFlag.ItemIsDropEnabled
        return flags

    def data(self, index: QModelIndex, role: int):
        if not index.isValid():
            return
        item: KeyValueTreeItem = self.itemFromIndex(index)
        if role == Qt.ItemDataRole.DisplayRole or role == Qt.ItemDataRole.EditRole:
            if index.column() == 0:
                return item.key()
            elif index.column() == 1:
                if item.isLeaf():
                    value = item.value()
                    if isinstance(value, np.ndarray):
                        value = ndarray_to_tuple(value)
                    else:
                        try:
                            # a numpy value
                            dtype = value.dtype
                            if np.issubdtype(dtype, np.floating):
                                value = float(value)
                            elif np.issubdtype(dtype, np.integer):
                                value = int(value)
                        except:
                            # not a numpy value
                            pass
                    return value
        elif role == Qt.ItemDataRole.DecorationRole:
            if index.column() == 0:
                if item.isLeaf():
                    return
                value = item.value()
                if isinstance(value, dict):
                    return qta.icon('ph.folder-thin')
                elif isinstance(value, list):
                    return qta.icon('ph.list-numbers-thin')

    def setData(self, index: QModelIndex, value, role: int) -> bool:
        if not index.isValid():
            return False
        item: KeyValueTreeItem = self.itemFromIndex(index)
        if role == Qt.ItemDataRole.EditRole:
            if index.column() == 0:
                item.setKey(value)
                self.dataChanged.emit(index, index)
                return True
            elif index.column() == 1:
                if item.isLeaf():
                    # overwriting a basic value, not a key:value map
                    if type(value) not in [list, dict]:
                        # no change to tree structure
                        item.setValue(value)
                        self.dataChanged.emit(index, index)
                        return True
                # change to tree structure
                # overwriting a key:value map
                self.beginResetModel()
                item.setValue(value)
                self.endResetModel()
                return True
        return False


def ndarray_to_tuple(arr: np.ndarray):
    if arr.shape == ():
        return arr.item()
    else:
        return tuple(map(ndarray_to_tuple, arr))



""" Tree view for an `AbstractTreeModel` with drag-and-drop, context menu, and mouse wheel expand/collapse.
"""

from __future__ import annotations
from typing import Callable, Any
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
import qtawesome as qta
from pyqt_ext.tree import AbstractTreeItem, AbstractTreeModel, AbstractTreeMimeData


class TreeView(QTreeView):

    selectionWasChanged = Signal()

    def __init__(self, *args, **kwargs) -> None:
        QTreeView.__init__(self, *args, **kwargs)

        # general settings
        sizePolicy = self.sizePolicy()
        sizePolicy.setHorizontalPolicy(QSizePolicy.Policy.Expanding)
        sizePolicy.setVerticalPolicy(QSizePolicy.Policy.Expanding)
        self.setSizePolicy(sizePolicy)
        self.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)

        # selection
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)

        # drag-n-drop
        self.setDragAndDropEnabled(True)

        # context menu defined in customContextMenu()
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self.onCustomContextMenuRequested)
    
    def refresh(self) -> None:
        self.storeState()
        self.model().reset()
        self.restoreState()
    
    def model(self) -> AbstractTreeModel:
        model: AbstractTreeModel = super().model()
        return model
    
    def treeData(self) -> Any:
        return self.model().treeData()
    
    def setTreeData(self, data: Any) -> None:
        self.storeState()
        if self.model() is None:
            self.setModel(AbstractTreeModel(data))
        else:
            self.model().setTreeData(data)
        self.restoreState()
    
    def storeState(self, items: list[AbstractTreeItem] = None) -> None:
        model: AbstractTreeModel = self.model()
        if model is None:
            return
        root: AbstractTreeItem = model.rootItem()
        if root is None:
            return
        if items is None:
            items = list(root.depthFirst())
        
        if not hasattr(self, '_view_state_'):
            self._view_state_ = {}
        
        selected_indexes: list[QModelIndex] = self.selectionModel().selectedIndexes()
        for item in items:
            if item is root:
                continue
            index: QModelIndex = model.indexFromItem(item)
            if not index.isValid():
                continue
            item._expanded_ = self.isExpanded(index)
            item._selected_ = index in selected_indexes
            self._view_state_[item.path()] = {
                'expanded': item._expanded_,
                'selected': item._selected_,
            }

    def restoreState(self, items: list[AbstractTreeItem] = None) -> None:
        model: AbstractTreeModel = self.model()
        if model is None:
            return
        root: AbstractTreeItem = model.rootItem()
        if root is None:
            return
        if items is None:
            items = list(root.depthFirst())
        
        state = getattr(self, '_view_state_', {})

        to_select: QItemSelection = QItemSelection()
        to_deselect: QItemSelection = QItemSelection()
        for item in items:
            if item is root:
                continue
            index: QModelIndex = model.indexFromItem(item)
            if not index.isValid():
                continue
            isExpanded = getattr(item, '_expanded_', None)
            if isExpanded is None:
                try:
                    isExpanded = state[item.path()]['expanded']
                except KeyError:
                    pass
            if isExpanded is not None:
                self.setExpanded(index, isExpanded)
            isSelected = getattr(item, '_selected_', None)
            if isSelected is None:
                try:
                    isSelected = state[item.path()]['selected']
                except KeyError:
                    pass
            if isSelected is not None:
                # flags = QItemSelectionModel.SelectionFlag.Rows
                # if isSelected:
                #     flags |= QItemSelectionModel.SelectionFlag.Select
                # else:
                #     flags |= QItemSelectionModel.SelectionFlag.Deselect
                # self.selectionModel().selection().select(index, index, flags)
                if isSelected:
                    to_select.merge(QItemSelection(index, index), QItemSelectionModel.SelectionFlag.Select | QItemSelectionModel.SelectionFlag.Rows)
                else:
                    to_deselect.merge(QItemSelection(index, index), QItemSelectionModel.SelectionFlag.Select | QItemSelectionModel.SelectionFlag.Rows)
        if to_deselect.count():
            self.selectionModel().select(to_deselect, QItemSelectionModel.SelectionFlag.Deselect | QItemSelectionModel.SelectionFlag.Rows)
        if to_select.count():
            self.selectionModel().select(to_select, QItemSelectionModel.SelectionFlag.Select | QItemSelectionModel.SelectionFlag.Rows)
    
    @Slot(QItemSelection, QItemSelection)
    def selectionChanged(self, selected: QItemSelection, deselected: QItemSelection) -> None:
        QTreeView.selectionChanged(self, selected, deselected)
        self.selectionWasChanged.emit()

    def selectedItems(self) -> list[AbstractTreeItem]:
        model: AbstractTreeModel = self.model()
        if model is None:
            return []
        indexes: list[QModelIndex] = self.selectionModel().selectedIndexes()
        items: list[AbstractTreeItem] = [model.itemFromIndex(index) for index in indexes]
        return items
    
    def setSelectedItems(self, items: list[AbstractTreeItem]):
        model: AbstractTreeModel = self.model()
        if model is None:
            return
        self.selectionModel().clearSelection()
        selection: QItemSelection = QItemSelection()
        flags = QItemSelectionModel.SelectionFlag.Select | QItemSelectionModel.SelectionFlag.Rows
        for item in items:
            index: QModelIndex = model.indexFromItem(item)
            if not index.isValid():
                continue
            selection.merge(QItemSelection(index, index), flags)
        if selection.count():
            self.selectionModel().select(selection, flags)
    
    def removeItems(self, items: list[AbstractTreeItem]) -> None:
        if not items:
            return
        self.model().removeItems(items)
    
    def removeSelectedItems(self) -> None:
        items: list[AbstractTreeItem] = self.selectedItems()
        self.removeItems(items)
    
    def askToRemoveItems(self, items: list[AbstractTreeItem], title: str = 'Remove', text: str = None) -> None:
        model: AbstractTreeModel = self.model()
        if model is None:
            return
        if not items:
            return
        if text is None:
            if len(items) == 1:
                text = f'Remove {items[0].path()}?'
            else:
                count: int = len(items)
                text = f'Remove {count} items including {items[0].path()}?'
        answer = QMessageBox.question(self, title, text, 
            buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No, 
            defaultButton=QMessageBox.StandardButton.No
        )
        if answer!= QMessageBox.StandardButton.Yes:
            return
        self.removeItems(items)
    
    def askToRemoveSelectedItems(self) -> None:
        items: list[AbstractTreeItem] = self.selectedItems()
        self.askToRemoveItems(items, text='Remove selected?')
    
    @Slot(QPoint)
    def onCustomContextMenuRequested(self, point: QPoint) -> None:
        index: QModelIndex = self.indexAt(point)
        menu: QMenu | None = self.customContextMenu(index)
        if menu is not None:
            menu.exec(self.viewport().mapToGlobal(point))
    
    def customContextMenu(self, index: QModelIndex = QModelIndex()) -> QMenu | None:
        model: AbstractTreeModel = self.model()
        if model is None:
            return
        
        menu = QMenu(self)

        # context menu for item that was clicked on
        if index.isValid():
            item: AbstractTreeItem = model.itemFromIndex(index)
            item_label = self.truncateLabel(item.path())
            menu.addAction(f'Remove {item_label}', lambda item=item: self.askToRemoveItems([item]))
            menu.addSeparator()
        
        self.appendDefaultContextMenu(menu)
        return menu
    
    def appendDefaultContextMenu(self, menu: QMenu) -> None:
        """ Append the default context menu items to the input menu.
        """
        menu.addAction('Expand All', self.expandAll)
        menu.addAction('Collapse All', self.collapseAll)
        if self.model().columnCount() > 1:
            menu.addAction('Resize Columns to Contents', self.resizeAllColumnsToContents)
            menu.addAction('Show All', self.showAll)
        
        if self.selectionMode() in [QAbstractItemView.SelectionMode.ContiguousSelection, QAbstractItemView.SelectionMode.ExtendedSelection, QAbstractItemView.SelectionMode.MultiSelection]:
            menu.addSeparator()
            menu.addAction('Select All', self.selectAll)
            menu.addAction('Select None', self.clearSelection)
        
        if len(self.selectedPaths()) > 1:
            menu.addSeparator()
            menu.addAction('Remove Selected', self.askToRemoveSelectedItems)

        menu.addSeparator()
        menu.addAction('Refresh', self.refresh)
    
    def truncateLabel(self, label: str, max_length: int = 50) -> str:
        """ Truncate long strings from the beginning.
        
        e.g., '...<end of label>'
        This preserves the final part of any tree paths used as labels.
        """
        if len(label) <= max_length:
            return label
        return '...' + label[-(max_length - 3):]
    
    def expandAll(self) -> None:
        QTreeView.expandAll(self)
        # store current expanded depth
        self._expanded_depth_ = self.model().maxDepth()
    
    def collapseAll(self) -> None:
        QTreeView.collapseAll(self)
        self._expanded_depth_ = 0
    
    def expandToDepth(self, depth: int) -> None:
        depth = max(0, min(depth, self.model().maxDepth()))
        if depth == 0:
            self.collapseAll()
            return
        QTreeView.expandToDepth(self, depth - 1)
        self._expanded_depth_ = depth
    
    def resizeAllColumnsToContents(self) -> None:
        for col in range(self.model().columnCount()):
            self.resizeColumnToContents(col)
    
    def showAll(self) -> None:
        self.expandAll()
        self.resizeAllColumnsToContents()
    
    def eventFilter(self, obj: QObject, event: QEvent) -> None:
        if event.type() == QEvent.Type.Wheel:
            modifiers: Qt.KeyboardModifier = event.modifiers()
            if Qt.KeyboardModifier.ControlModifier in modifiers \
            or Qt.KeyboardModifier.AltModifier in modifiers \
            or Qt.KeyboardModifier.MetaModifier in modifiers:
                self.mouseWheelEvent(event)
                return True
        # elif event.type() == QEvent.Type.FocusIn:
        #     print('FocusIn')
        #     # self.showDropIndicator(True)
        # elif event.type() == QEvent.Type.FocusOut:
        #     print('FocusOut')
        #     # self.showDropIndicator(False)
        return QTreeView.eventFilter(self, obj, event)
    
    def mouseWheelEvent(self, event: QWheelEvent) -> None:
        delta: int = event.angleDelta().y()
        depth = getattr(self, '_expanded_depth_', 0)
        if delta > 0:
            self.expandToDepth(depth + 1)
        elif delta < 0:
            self.expandToDepth(depth - 1)
    
    def setDragAndDropEnabled(self, enabled: bool) -> None:
        if enabled:
            self.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
            self.setDefaultDropAction(Qt.DropAction.MoveAction)
        else:
            self.setDragDropMode(QAbstractItemView.DragDropMode.NoDragDrop)
        self.setDragEnabled(enabled)
        self.setAcceptDrops(enabled)
        self.viewport().setAcceptDrops(enabled)
        self.setDropIndicatorShown(enabled)
    
    def dragEnterEvent(self, event: QDragEnterEvent) -> None:
        mime_data = event.mimeData()
        if isinstance(mime_data, AbstractTreeMimeData) and (mime_data.model is self.model()):
            # Store the current state of the dragged paths and all their descendent paths in the MIME data.
            # We only want to do this for the model where the drag was initiated (i.e., mime_data.model).
            # We'll use this stored state in the dropEvent to restore the view of the dropped items.
            items: list[AbstractTreeItem] = mime_data.items.copy()
            for root_item in mime_data.items:
                for item in root_item.depthFirst():
                    if item not in items:
                        items.append(item)
            self.storeState(items)
        QTreeView.dragEnterEvent(self, event)
    
    def dropEvent(self, event: QDropEvent) -> None:
        mime_data = event.mimeData()
        QTreeView.dropEvent(self, event)
        if isinstance(mime_data, AbstractTreeMimeData):
            # update state of dragged items and all their descendents as specified in the MIME data
            items: list[AbstractTreeItem] = mime_data.items.copy()
            for root_item in mime_data.items:
                for item in root_item.depthFirst():
                    if item not in items:
                        items.append(item)
            self.restoreState(items)
    
    """
    Everything below here is for path-based manipulation of the tree.
    !! This typically only makes sense if all sibling items have unique names.
       If sibling items should be able to have the same name, you probably should not use the functions below.
    """
    
    def selectedPaths(self) -> list[str]:
        items: list[AbstractTreeItem] = self.selectedItems()
        paths: list[str] = [item.path() for item in items]
        return paths
    
    def setSelectedPaths(self, paths: list[str]):
        model: AbstractTreeModel = self.model()
        if model is None:
            return
        items: list[AbstractTreeItem] = [model.itemFromPath(path) for path in paths]
        self.setSelectedItems(items)
    
    def removePaths(self, paths: list[str]) -> None:
        model: AbstractTreeModel = self.model()
        if model is None:
            return
        items: list[AbstractTreeItem] = [model.itemFromPath(path) for path in paths]
        self.removeItems(items)
    
    def askToRemovePaths(self, paths: list[str], title: str = 'Remove', text: str = None) -> None:
        model: AbstractTreeModel = self.model()
        if model is None:
            return
        items: list[AbstractTreeItem] = [model.itemFromPath(path) for path in paths]
        self.askToRemoveItems(items, title, text)
    
    def askToRemoveSelectedPaths(self) -> None:
        paths: list[str] = self.selectedPaths()
        self.askToRemovePaths(paths, text='Remove selected?')


def test_live():
    root1 = AbstractTreeItem()
    AbstractTreeItem(parent=root1)
    root1.appendChild(AbstractTreeItem(name='child2'))
    root1.insertChild(1, AbstractTreeItem(name='child3'))
    root1.children[1].appendChild(AbstractTreeItem())
    root1.children[1].appendChild(AbstractTreeItem())
    root1.children[1].appendChild(AbstractTreeItem())
    grandchild2 = AbstractTreeItem(name='grandchild2')
    grandchild2.setParent(root1['child2'])
    AbstractTreeItem(name='greatgrandchild', parent=root1['/child2/grandchild2'])

    root2 = AbstractTreeItem()
    AbstractTreeItem(parent=root2)
    root2.appendChild(AbstractTreeItem(name='child22'))
    root2.insertChild(1, AbstractTreeItem(name='child23'))
    root2.children[1].appendChild(AbstractTreeItem())
    root2.children[1].appendChild(AbstractTreeItem())
    root2.children[1].appendChild(AbstractTreeItem())
    grandchild2 = AbstractTreeItem(name='grandchild22')
    grandchild2.setParent(root2['child22'])
    AbstractTreeItem(name='greatgrandchild2', parent=root2['/child22/grandchild22'])

    app = QApplication()
    model1 = AbstractTreeModel(root1)
    view1 = TreeView()
    view1.setModel(model1)
    view1.show()
    view1.resize(QSize(800, 600))
    view1.showAll()
    view1.setWindowTitle('TreeView 1')
    model2 = AbstractTreeModel(root2)
    view2 = TreeView()
    view2.setModel(model2)
    view2.show()
    view2.resize(QSize(800, 600))
    view2.showAll()
    view2.setWindowTitle('TreeView 2')
    app.exec()
    print(root1)
    print(root2)


if __name__ == '__main__':
    test_live()



""" Base class for a tree of items to be used as a data interface with AbstractTreeModel(QAbstractItemModel).
"""

from __future__ import annotations
from warnings import warn
from typing import Callable
from collections.abc import Iterator


class AbstractTreeItem():
    """ Base class for a tree of items to be used as a data interface with AbstractTreeModel(QAbstractItemModel).
    
    !!! This only implements the tree structure.
        You must define any attributes to interface with the actual tree data in a derived class.
    
    What you get out-of-the-box:
        - A host of methods for navigating and manipulating the tree structure:
            - Parent/child linkage.
            - Host of other linkage properties: siblings, first/last child, first/last sibling, etc.
            - Path-based item access based on item names.
            - Depth-first, leaves, and ancestors iteration.
            - String representation of the tree.
            - And more...
        - An interface that is designed to work with AbstractTreeModel (i.e., QAbstractItemModel).

    In a derived class:
        1. Add attributes to store/interface with your tree data.
        2. Reimplement methods `setParent` and `insertChild` if you need to edit the structure of the underlying tree data (all other tree manipulation methods are based on these two methods). Out-of-the-box these methods only alter the structure of the tree item wrappers, not the underlying data which this abstract class knows nothing about. Note: If each tree item owns its own data, then you probably don't need to touch these.
        3. For nice printout, you may want to reimplement `__repr__`.
    
    For an example, see `KeyValueTreeItem`.
    """
    
    def __init__(self, name: str | None = None, parent: AbstractTreeItem | None = None) -> None:
        self._name: str | None = name
        self._parent: AbstractTreeItem | None = None
        self.children: list[AbstractTreeItem] = []
        if parent is not None:
            # properly handle parent linkage
            self.setParent(parent)
    
    # pretty print tree --------------------------------------------------

    def __repr__(self) -> str:
        """ Return a single line string representation of this item.
        
        See __str__ for a multi-line representation of the tree.
        """
        name: str = self.name()
        parent_name: str = self.parent().name() if self.parent() else '/'
        child_names: str = ','.join([child.name() for child in self.children])
        return f'name: {name}; parent: {parent_name}; children: {child_names}'
    
    def __str__(self) -> str:
        """ Returns a multi-line string representation of this item's tree branch.

        Each item is described by its name.
        """
        return self._tree_repr(lambda item: item.name())
    
    def dumps(self) -> str:
        """ Returns a multi-line string representation of this item's tree branch.

        Each item is described by its repr.
        """
        return self._tree_repr(lambda item: repr(item))
    
    def _tree_repr(self, func: Callable[[AbstractTreeItem], str]) -> str:
        """ Returns a multi-line string representation of this item's tree branch.

        Each item is described by the single line str returned by func(item).
        See __str__ and dumps for examples.
        """
        items: list[AbstractTreeItem] = list(self.depthFirst())
        lines: list[str] = [func(item) for item in items]
        for i, item in enumerate(items):
            if item is self:
                continue
            if item is item.parent().lastChild():
                lines[i] = '\u2514' + '\u2500'*2 + ' ' + lines[i]
            else:
                lines[i] = '\u251C' + '\u2500'*2 + ' ' + lines[i]
            parent = item.parent()
            while parent is not self:
                if i < items.index(parent.parent().lastChild()):
                    lines[i] = '\u2502' + ' '*3 + lines[i]
                else:
                    lines[i] = ' '*4 + lines[i]
                parent = parent.parent()
        return '\n'.join(lines)
    
    # path-based item access based on item name --------------------------------------------------

    def __getitem__(self, path: str) -> AbstractTreeItem:
        """ Return subtree item at path starting from this item.

        !! For unique item access, all paths in the tree must be unique.
           Unique paths are not a requirement, it is up to you to enforce this if you want it.
           If the path is not unique, the first item with path is returned.
        """
        item: AbstractTreeItem = self
        path_parts = path.strip('/').split('/')
        for name in path_parts:
            try:
                child_names = [child.name() for child in item.children]
                child_index = child_names.index(name)
                item = item.children[child_index]
                if child_names.count(name) > 1:
                    warn('Path is not unique.')
            except Exception as error:
                warn(str(error))
                return None
        return item
    
    def __setitem__(self, path: str, new_item: AbstractTreeItem) -> None:
        """ Set subtree item at path starting from this item.

        !! For unique item access, all paths in the tree must be unique.
           Unique paths are not a requirement, it is up to you to enforce this if you want it.
           If the path is not unique, the first item with path will be set to the new item.
        """
        item: AbstractTreeItem = self
        path_parts = path.strip('/').split('/')
        if len(path_parts) == 0:
            raise ValueError('An item cannot set itself to a new item.')
        for name in path_parts[:-1]:
            try:
                child_names = [child.name() for child in item.children]
                child_index = child_names.index(name)
                item = item.children[child_index]
                if child_names.count(name) > 1:
                    warn('Path is not unique.')
            except Exception as error:
                # create new tree item to ensure validity of path
                item = AbstractTreeItem(name=name, parent=item)
        # set new_item at path
        new_item_name = path_parts[-1]
        child_names = [child.name() for child in item.children]
        if new_item_name in child_names:
            # replace item at path with new_item
            child_index = child_names.index(new_item_name)
            item.children[child_index].setParent(None)  # remove current item at path
            item.insertChild(child_index, new_item)  # insert new_item at path
        else:
            # add new_item at path
            new_item.setParent(item)
        # name new_item according to path (ignore's new_item's current name)
        new_item.setName(new_item_name)
    
    def name(self) -> str:
        """ Retun this item's name, or the item's sibling index (or '' if root) if name is not defined.
        """
        if (self._name is None) or (self._name == ''):
            # return f'{self.__class__.__name__}@{id(self)}'
            if self.parent() is None:
                return ''
            row = self.siblingIndex()
            return f'{row}'
        return self._name
    
    def setName(self, name: str) -> None:
        self._name = name
    
    # tree linkage --------------------------------------------------

    def parent(self) -> AbstractTreeItem | None:
        return self._parent
    
    def setParent(self, parent: AbstractTreeItem | None) -> None:
        """ This is the primary method for restructuring the tree.

        This method restructures the AbstractTreeItem interface only,
        not any data associated with the AbstractTreeItems.

        Derived classes that need to mirror changes in the tree structure to their data 
        should reimplement this method.
        
        Note: Together, setParent and insertChild cover all needed tree restructuring.
        See appendChild, insertChild, and removeChild.
        """
        oldParent: AbstractTreeItem | None = self.parent()
        newParent: AbstractTreeItem | None = parent
        if oldParent is newParent:
            # nothing to do
            return
        if (newParent is not None) and newParent.hasAncestor(self):
            raise ValueError('Cannot set parent to a descendant.')
        
        if oldParent is not None:
            # detach from old parent
            if self in oldParent.children:
                oldParent.children.remove(self)
        if newParent is not None:
            # attach to new parent (appends as last child)
            if self not in newParent.children:
                newParent.children.append(self)
        self._parent = newParent
    
    def path(self) -> str:
        """ Return the /path/to/this/item from the root item.

        Path is constructed from item names separated by /.
        """
        item: AbstractTreeItem = self
        path = []
        while not item.isRoot():
            path.insert(0, item.name())
            item = item.parent()
        return '/' + '/'.join(path)
    
    def root(self) -> AbstractTreeItem:
        """ Return the root item of this item's branch. """
        item: AbstractTreeItem = self
        while item.parent() is not None:
            item = item.parent()
        return item
    
    def firstChild(self) -> AbstractTreeItem | None:
        if self.children:
            return self.children[0]

    def lastChild(self) -> AbstractTreeItem | None:
        if self.children:
            return self.children[-1]

    def siblings(self) -> list[AbstractTreeItem]:
        """ Return a list of this item's siblings (inclusive of this item).

        Return a copy so that manipulating the returned list does not affect the tree structure.
        """
        if self.parent() is not None:
            return self.parent().children.copy()
        return [self]
    
    def firstSibling(self) -> AbstractTreeItem:
        """ Note: Can be this item.
        """
        if self.parent() is not None:
            return self.parent().firstChild()
        return self

    def lastSibling(self) -> AbstractTreeItem:
        """ Note: Can be this item.
        """
        if self.parent() is not None:
            return self.parent().lastChild()
        return self

    def nextSibling(self) -> AbstractTreeItem | None:
        if self.parent() is not None:
            siblings: list[AbstractTreeItem] = self.parent().children
            i: int = siblings.index(self)
            if i+1 < len(siblings):
                return siblings[i+1]

    def prevSibling(self) -> AbstractTreeItem | None:
        if self.parent() is not None:
            siblings: list[AbstractTreeItem] = self.parent().children
            i: int = siblings.index(self)
            if i-1 >= 0:
                return siblings[i-1]

    def siblingIndex(self) -> int | None:
        if self.parent() is not None:
            return self.parent().children.index(self)
    
    def isRoot(self) -> bool:
        return self.parent() is None
    
    def isLeaf(self) -> bool:
        return not self.children
    
    def hasAncestor(self, ancestor: AbstractTreeItem) -> bool:
        item: AbstractTreeItem = self.parent()
        while item is not None:
            if item is ancestor:
                return True
            item = item.parent()
        return False
    
    # tree mutation --------------------------------------------------

    def appendChild(self, item: AbstractTreeItem) -> None:
        """ Reparent item as this item's last child.
        """
        item.setParent(self)
    
    def insertChild(self, index: int, item: AbstractTreeItem) -> None:
        """ Reparent item as this item's index-th child.

        If you need to update the underlying data structure, you must reimplement both setParent() and this method in a derived class.
        """
        if not (0 <= index <= len(self.children)):
            raise IndexError('Index out of range.')
        # append as last child
        item.setParent(self)
        # move item to index
        pos = self.children.index(item)
        if pos != index:
            if pos < index:
                index -= 1
            if pos != index:
                self.children.insert(index, self.children.pop(pos))
    
    def removeChild(self, item: AbstractTreeItem) -> None:
        """ Remove child from this item's children list and orphan it.
        """
        if item.parent() is not self:
            raise ValueError('Item is not a child of this item.')
        item.setParent(None)
    
    # ancestors iteration --------------------------------------------------

    def parents(self) -> Iterator[AbstractTreeItem]:
        item: AbstractTreeItem | None = self.parent()
        end_item = None
        while item is not end_item:
            yield item
            item = item.parent()
    
    # depth-first iteration --------------------------------------------------
    
    def depthFirst(self) -> Iterator[AbstractTreeItem]:
        item: AbstractTreeItem = self
        end_item: AbstractTreeItem | None = self.lastDepthFirst().nextDepthFirst()
        while item is not end_item:
            yield item
            item = item.nextDepthFirst()
    
    def reverseDepthFirst(self) -> Iterator[AbstractTreeItem]:
        item: AbstractTreeItem = self.lastDepthFirst()
        end_item: AbstractTreeItem | None = self.prevDepthFirst()
        while item is not end_item:
            yield item
            item = item.prevDepthFirst()
    
    def nextDepthFirst(self) -> AbstractTreeItem | None:
        if self.children:
            return self.firstChild()
        next_sibling: AbstractTreeItem = self.nextSibling()
        if next_sibling is not None:
            return next_sibling
        item: AbstractTreeItem = self.parent()
        while item is not None:
            next_sibling: AbstractTreeItem = item.nextSibling()
            if next_sibling is not None:
                return next_sibling
            item = item.parent()
        return None

    def prevDepthFirst(self) -> AbstractTreeItem | None:
        prev_sibling: AbstractTreeItem = self.prevSibling()
        if prev_sibling is not None:
            return prev_sibling.lastDepthFirst()
        if self.parent() is not None:
            return self.parent()
        return None
    
    def lastDepthFirst(self) -> AbstractTreeItem:
        item: AbstractTreeItem = self
        while item.children:
            item = item.lastChild()
        return item
    
    # leaf iteration --------------------------------------------------
    
    def leaves(self) -> Iterator[AbstractTreeItem]:
        item: AbstractTreeItem = self.firstLeaf()
        end_item: AbstractTreeItem | None = self.lastLeaf().nextLeaf()
        while item is not end_item:
            yield item
            item = item.nextLeaf()
    
    def reverseLeaves(self) -> Iterator[AbstractTreeItem]:
        item: AbstractTreeItem = self.lastLeaf()
        end_item: AbstractTreeItem | None = self.firstLeaf().prevLeaf()
        while item is not end_item:
            yield item
            item = item.prevLeaf()
    
    def nextLeaf(self) -> AbstractTreeItem | None:
        try:
            return self.nextDepthFirst().firstLeaf()
        except Exception:
            return None

    def prevLeaf(self) -> AbstractTreeItem | None:
        item: AbstractTreeItem | None = self.prevDepthFirst()
        while (item is not None) and item.children:
            item = item.prevDepthFirst()
        return item
    
    def firstLeaf(self) -> AbstractTreeItem:
        item: AbstractTreeItem = self
        while item.children:
            item = item.firstChild()
        return item
    
    def lastLeaf(self) -> AbstractTreeItem:
        item: AbstractTreeItem = self
        while item.children:
            item = item.lastChild()
        return item

    # misc --------------------------------------------------

    def depth(self, root: AbstractTreeItem = None) -> int:
        """ Return depth of this item in the branch starting at root (or entire tree if root is None).
        """
        depth: int = 0
        item: AbstractTreeItem = self
        while (item.parent() is not None) and (item is not root):
            depth += 1
            item = item.parent()
        return depth
    
    def maxDepthBelow(self) -> int:
        # Return the maximum depth within this item's branch.
        max_depth: int = 0
        for item in self.leaves():
            depth: int = item.depth(self)
            if depth > max_depth:
                max_depth = depth
        return max_depth
    
    @staticmethod
    def uniqueName(name: str, names: list[str], unique_counter_start: int = 2) -> str:
        """ May be useful for trees that require unique paths (i.e., sibling names).

        If name in names, return name_2, or name_3, etc. until a unique name is found.
        """
        if name not in names:
            return name
        base_name = name
        i = unique_counter_start
        name = f'{base_name}_{i}'
        while name in names:
            i += 1
            name = f'{base_name}_{i}'
        return name


def test_tree():
    root = AbstractTreeItem()
    AbstractTreeItem(parent=root)
    root.appendChild(AbstractTreeItem(name='child2'))
    root.insertChild(1, AbstractTreeItem(name='child3'))
    root.children[1].appendChild(AbstractTreeItem())
    root.children[1].appendChild(AbstractTreeItem())
    root.children[1].appendChild(AbstractTreeItem())
    grandchild2 = AbstractTreeItem(name='grandchild2')
    grandchild2.setParent(root['child2'])
    AbstractTreeItem(name='greatgrandchild', parent=root['child2/grandchild2'])
    
    print('\nInitial tree...')
    print(root)

    print('\nDepth first iteration...')
    for item in root.depthFirst():
        print(item.name())

    print('\nRemove grandchild2...')
    grandchild2.setParent(None)
    print(root)

    print('\nInsert grandchild2...')
    grandchild2.setParent(root['child2'])
    print(root)


if __name__ == '__main__':
    test_tree()



""" PySide/PyQt table widget with copy/paste.
"""

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QTableWidget, QApplication, QTableWidgetItem


class TableWidgetWithCopyPaste(QTableWidget):
    """ QTableWidget with copy/paste to/from clipboard in CSV format.
    """
    
    def __init__(self, *args, **kwargs):
        QTableWidget.__init__(self, *args, **kwargs)
    
    def keyPressEvent(self, event):
        QTableWidget.keyPressEvent(self, event)

        if event.key() == Qt.Key.Key_C and (event.modifiers() & Qt.KeyboardModifier.ControlModifier):
            self.copy_selected_cells()
        elif event.key() == Qt.Key.Key_V and (event.modifiers() & Qt.KeyboardModifier.ControlModifier):
            self.paste_to_cells()
    
    def copy_selected_cells(self):
        copied_cells = sorted(self.selectedIndexes())
        copy_text = ''
        max_column = copied_cells[-1].column()
        max_row = copied_cells[-1].row()
        for c in copied_cells:
            copy_text += self.item(c.row(), c.column()).text()
            if c.column() == max_column:
                if c.row() != max_row:
                    copy_text += '\n'
            else:
                copy_text += '\t'
        QApplication.clipboard().setText(copy_text)
    
    def paste_to_cells(self):
        selection = self.selectedIndexes()
        if selection:
            row_anchor = selection[0].row()
            column_anchor = selection[0].column()
            clipboard = QApplication.clipboard()
            rows = clipboard.text().split('\n')
            for indx_row, row in enumerate(rows):
                values = row.split('\t')
                for indx_col, value in enumerate(values):
                    if row_anchor + indx_row < self.rowCount() and column_anchor + indx_col < self.columnCount():
                        item = QTableWidgetItem(value)
                        self.setItem(row_anchor + indx_row, column_anchor + indx_col, item)


def test_live():
    app = QApplication()

    table = TableWidgetWithCopyPaste(3, 3)
    for i in range(3):
        for j in range(3):
            table.setItem(i, j, QTableWidgetItem(f'{i},{j}'))
    table.show()

    app.exec()


if __name__ == '__main__':
    test_live()



""" LinearRegionItem with context menu, optional text label, and style dialog.
"""

from __future__ import annotations
from qtpy.QtCore import *
from qtpy.QtGui import *
from qtpy.QtWidgets import *
from pyqt_ext.utils import toColorStr, toQColor
from pyqt_ext.widgets import ColorButton, CollapsibleSection
import pyqtgraph as pg


class AxisRegion(pg.LinearRegionItem):
    """ LinearRegionItem with context menu, optional text label, and style dialog.
    
    self.sigRegionChangeFinished is emitted when the item is moved or resized.
    """

    sigRegionDragFinished = Signal(object)
    sigEditingFinished = Signal(object)
    sigDeletionRequested = Signal(object)

    def __init__(self, *args, **kwargs):
        if 'orientation' not in kwargs:
            kwargs['orientation'] = 'vertical'
        if 'brush' not in kwargs:
            kwargs['brush'] = pg.mkBrush(QColor(237, 135, 131, 51))
        if 'hoverBrush' not in kwargs:
            kwargs['hoverBrush'] = pg.mkBrush(QColor(237, 135, 131, 128))
        if 'pen' not in kwargs:
            kwargs['pen'] = pg.mkPen(QColor(237, 135, 131), width=1)
        if 'hoverPen' not in kwargs:
            kwargs['hoverPen'] = pg.mkPen(QColor(255, 0, 0), width=2)
        if 'swapMode' not in kwargs:
            kwargs['swapMode'] = 'push'  # keeps label on left side
        pg.LinearRegionItem.__init__(self, *args, **kwargs)

        self._textLabelItem: pg.InfLineLabel = pg.InfLineLabel(self.lines[0], text='', movable=True, position=1, anchors=[(0,0), (0,0)])
        self._textLabelItem.setVisible(False)
        self.setFontColor(QColor.fromRgbF(0.15, 0.15, 0.15))

        self.lines[0].sigClicked.connect(self.onEdgeClicked)
        self.lines[1].sigClicked.connect(self.onEdgeClicked)

        # update label position when region is moved or resized
        # TODO: disallow dragging label outside of viewbox
        self.sigRegionChanged.connect(self.updateLabelPosition)
        # self.sigRegionChangeFinished.connect(lambda self=self: self.storeState())

        self.setZValue(11)

        self.contextMenu = QMenu(self.__class__.__name__)
        self.contextMenu.addAction('Edit Region', lambda: self.editDialog(self.getViewWidget()))
        self.contextMenu.addSeparator()
        self._movableAction = self.contextMenu.addAction('Movable', lambda: self.toggleMovable())
        self._movableAction.setCheckable(True)
        self._movableAction.setChecked(self.movable)
        self.contextMenu.addSeparator()
        self.contextMenu.addAction('Delete Region', lambda: self.sigDeletionRequested.emit(self))

        self.includeParentContextMenus = False
    
    def getState(self) -> dict:
        """ Return hashable dict for saving and restoring state.
        """
        state = {
            'position': self.getRegion(),
            'movable': self.movable,
            'text': self.text(),
            'format': self.getFormat(),
        }
        return state
    
    def setState(self, state: dict):
        """ Restore state from hashable dict.
        """
        for key, value in state.items():
            key = key.lower()
            if key == 'position':
                self.setRegion(value)
            elif key == 'movable':
                self.setMovable(value)
            elif key == 'text':
                self.setText(value)
            elif key == 'format':
                self.setFormat(value)
    
    def getFormat(self) -> dict:
        """ Return hashable dict for saving and restoring state.
        """
        return {
            'facecolor': toColorStr(self.faceColor()),
            'edgecolor': toColorStr(self.edgeColor()),
            'edgewidth': self.edgeWidth(),
            'facehovercolor': toColorStr(self.faceHoverColor()),
            'edgehovercolor': toColorStr(self.edgeHoverColor()),
            'edgehoverwidth': self.edgeHoverWidth(),
            'font': self.font().toString(),
            'fontsize': self.fontSize(),
            'fontcolor': toColorStr(self.fontColor()),
        }
    
    def setFormat(self, state: dict):
        """ Restore state from hashable dict.
        """
        for key, value in state.items():
            key = key.lower()
            if key == 'facecolor':
                self.setFaceColor(toQColor(value))
            elif key == 'edgecolor':
                self.setEdgeColor(toQColor(value))
            elif key == 'edgewidth':
                self.setEdgeWidth(value)
            elif key == 'facehovercolor':
                self.setFaceHoverColor(toQColor(value))
            elif key == 'edgehovercolor':
                self.setEdgeHoverColor(toQColor(value))
            elif key == 'edgehoverwidth':
                self.setEdgeHoverWidth(value)
            elif key == 'font':
                font = QFont()
                font.fromString(value)
                self.setFont(font)
            elif key == 'fontsize':
                self.setFontSize(value)
            elif key == 'fontcolor':
                self.setFontColor(toQColor(value))
    
    def storeState(self):
        self._state = self.getState()
    
    def restoreState(self):
        if not hasattr(self, '_state'):
            raise AttributeError('State has not been stored')
        self.setState(self._state)
    
    def setMovable(self, movable: bool):
        super().setMovable(movable)
        if hasattr(self, '_movableAction'):
            self._movableAction.setChecked(self.movable)
    
    def toggleMovable(self) -> None:
        self.setMovable(not self.movable)
        self.sigEditingFinished.emit(self)  # to update any data references
    
    def faceColor(self) -> QColor:
        return self.brush.color()
    
    def setFaceColor(self, color: QColor):
        self.brush.setColor(color)
    
    def edgePen(self) -> QPen:
        return self.lines[0].pen
    
    def setEdgePen(self, pen: QPen):
        self.lines[0].pen = pen
        self.lines[1].pen = pen
    
    def edgeColor(self) -> QColor:
        return self.lines[0].pen.color()
    
    def setEdgeColor(self, color: QColor):
        self.lines[0].pen.setColor(color)
        self.lines[1].pen.setColor(color)
    
    def edgeWidth(self) -> float:
        return self.lines[0].pen.width()
    
    def setEdgeWidth(self, width: float):
        self.lines[0].pen.setWidth(width)
        self.lines[1].pen.setWidth(width)
    
    def faceHoverColor(self) -> QColor:
        return self.hoverBrush.color()
    
    def setFaceHoverColor(self, color: QColor):
        self.hoverBrush.setColor(color)
    
    def edgeHoverPen(self) -> QPen:
        return self.lines[0].hoverPen
    
    def setEdgeHoverPen(self, pen: QPen):
        self.lines[0].hoverPen = pen
        self.lines[1].hoverPen = pen
    
    def edgeHoverColor(self) -> QColor:
        return self.lines[0].hoverPen.color()
    
    def setEdgeHoverColor(self, color: QColor):
        self.lines[0].hoverPen.setColor(color)
        self.lines[1].hoverPen.setColor(color)
    
    def edgeHoverWidth(self) -> float:
        return self.lines[0].hoverPen.width()
    
    def setEdgeHoverWidth(self, width: float):
        self.lines[0].hoverPen.setWidth(width)
        self.lines[1].hoverPen.setWidth(width)

    def text(self):
        try:
            return self._textLabelItem.format
        except:
            return ''

    def setText(self, text: str):
        self._textLabelItem.setFormat(text)
        self._textLabelItem.setVisible(text != '')
    
    def font(self) -> QFont:
        return self._textLabelItem.textItem.font()
    
    def setFont(self, font: QFont):
        self._textLabelItem.setFont(font)
    
    def fontSize(self) -> int:
        return self._textLabelItem.textItem.font().pointSize()
    
    def setFontSize(self, size):
        font = self._textLabelItem.textItem.font()
        font.setPointSize(size)
        self._textLabelItem.setFont(font)
    
    def fontColor(self) -> QColor:
        return self._textLabelItem.color
    
    def setFontColor(self, color: QColor):
        self._textLabelItem.setColor(color)
    
    def copyFormat(self, other: AxisRegion):
        self.setFormat(other.getFormat())
    
    def updateLabelPosition(self):
        if self._textLabelItem is not None:
            self._textLabelItem.updatePosition()
            pos = self._textLabelItem.orthoPos
            if pos < 0.05:
                self._textLabelItem.setPosition(0.05)
    
    def onEdgeClicked(self, line, event):
        if event.button() == Qt.RightButton:
            if self.raiseContextMenu(event):
                event.accept()
    
    def mouseClickEvent(self, event):
        if event.button() == Qt.MouseButton.RightButton:
            if self.boundingRect().contains(event.pos()):
                if self.raiseContextMenu(event):
                    event.accept()
                    return
        super().mouseClickEvent(event)
    
    def mouseDoubleClickEvent(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            self.editDialog(self.getViewWidget())
            event.accept()
            return
        super().mouseDoubleClickEvent(event)
    
    def mouseDragEvent(self, event):
        """ Add new signal for when drag is finished.
        """
        if not self.movable or event.button() != Qt.MouseButton.LeftButton:
            super().mouseDragEvent(event)
            return
        
        event.accept()
        
        if event.isStart():
            bdp = event.buttonDownPos()
            self.cursorOffsets = [l.pos() - bdp for l in self.lines]
            self.startPositions = [l.pos() for l in self.lines]
            self.moving = True
            
        if not self.moving:
            return
            
        self.blockLineSignal = True  # only want to update once
        for i, l in enumerate(self.lines):
            l.setPos(self.cursorOffsets[i] + event.pos())
        self.prepareGeometryChange()
        self.blockLineSignal = False
        
        if event.isFinish():
            self.moving = False
            self.sigRegionChangeFinished.emit(self)
            self.sigRegionDragFinished.emit(self)
        else:
            self.sigRegionChanged.emit(self)
    
    def raiseContextMenu(self, event):
        menu = self.getContextMenus(event)
        pos = event.screenPos()
        menu.popup(QPoint(int(pos.x()), int(pos.y())))
        return True
    
    def getContextMenus(self, event=None):
        self.menu = QMenu()
        # self.menu.addMenu(self.contextMenu)
        for action in self.contextMenu.actions():
            self.menu.addAction(action)

        if self.includeParentContextMenus:
        # Let the scene add on to the end of our context menu (this is optional)
            self.menu.addSection('View')
            self.menu = self.scene().addParentContextMenus(self, self.menu, event)
        
        return self.menu
    
    def editDialog(self, parent: QWidget = None):
        editAxisRegion(self, parent=parent)
        self.sigEditingFinished.emit(self)


class XAxisRegion(AxisRegion):
    """ Vertical AxisRegionItem for x-axis ROI. """

    def __init__(self, *args, **kwargs):
        kwargs['orientation'] = 'vertical'
        AxisRegion.__init__(self, *args, **kwargs)


class YAxisRegion(AxisRegion):
    """ Horizontal AxisRegionItem for y-axis ROI. """

    def __init__(self, *args, **kwargs):
        kwargs['orientation'] = 'horizontal'
        AxisRegion.__init__(self, *args, **kwargs)


class AxisRegionPanel(QWidget):

    def __init__(self, *args, **kwargs):
        QWidget.__init__(self, *args, **kwargs)

        form = QFormLayout(self)
        form.setContentsMargins(5, 5, 5, 5)
        form.setSpacing(5)
        form.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)

        # self._regionEdit = QLineEdit()
        # self._regionEdit.setToolTip('min [,max]')
        # self._regionEdit.setPlaceholderText('min [,max]')
        # form.addRow('Region', self._regionEdit)

        self._minEdit = QLineEdit()
        self._maxEdit = QLineEdit()
        form.addRow('Min', self._minEdit)
        form.addRow('Max', self._maxEdit)

        self._movableCheckBox = QCheckBox()
        self._movableCheckBox.setChecked(True)
        form.addRow('Movable', self._movableCheckBox)

        self._textEdit = QTextEdit()
        form.addRow('Text', self._textEdit)

        self._formatPanel = AxisRegionFormatPanel()
        self._formatPanelWrapperLayout = QVBoxLayout()
        self._formatPanelWrapperLayout.setContentsMargins(0, 0, 0, 0)
        self._formatPanelWrapperLayout.addWidget(self._formatPanel)

        self._formatSection = CollapsibleSection(title='Format')
        self._formatSection.setContentLayout(self._formatPanelWrapperLayout)
        form.addRow(self._formatSection)

        # default settings
        # self.setState(AxisRegion().getState())
    
    def getState(self):
        state = {}

        if self._minEdit.text() != '' and self._maxEdit.text() != '':
            state['position'] = tuple(sorted([float(self._minEdit.text()), float(self._maxEdit.text())]))

        if self._movableCheckBox.checkState() != Qt.CheckState.PartiallyChecked:
            state['movable'] = self._movableCheckBox.isChecked()
        
        state['text'] = self._textEdit.toPlainText()

        state['format'] = self._formatPanel.getFormat()

        return state

    def setState(self, state):
        for key, value in state.items():
            key = key.lower()
            if key == 'position':
                self._minEdit.setText(f'{value[0]:.6f}')
                self._maxEdit.setText(f'{value[1]:.6f}')
            elif key == 'movable':
                self._movableCheckBox.setChecked(value)
            elif key == 'text':
                self._textEdit.setPlainText(value)
            elif key == 'format':
                self._formatPanel.setFormat(value)


class AxisRegionFormatPanel(QWidget):

    def __init__(self, *args, **kwargs):
        QWidget.__init__(self, *args, **kwargs)

        self._faceColorButton = ColorButton()
        self._edgeColorButton = ColorButton()
        self._edgeWidthSpinBox = QDoubleSpinBox()

        self._faceHoverColorButton = ColorButton()
        self._edgeHoverColorButton = ColorButton()
        self._edgeHoverWidthSpinBox = QDoubleSpinBox()

        self._fontColorButton = ColorButton()
        self._fontSizeSpinBox = QDoubleSpinBox()

        self._formatLayout = QFormLayout()
        self._formatLayout.setContentsMargins(5, 5, 5, 5)
        self._formatLayout.setSpacing(5)
        self._formatLayout.addRow('Face Color', self._faceColorButton)
        self._formatLayout.addRow('Edge Color', self._edgeColorButton)
        self._formatLayout.addRow('Edge Width', self._edgeWidthSpinBox)
        self._formatLayout.addRow('Font Color', self._fontColorButton)
        self._formatLayout.addRow('Font Size', self._fontSizeSpinBox)

        self._hoverLayout = QFormLayout()
        self._hoverLayout.setContentsMargins(5, 5, 5, 5)
        self._hoverLayout.setSpacing(5)
        self._hoverLayout.addRow('Face Hover Color', self._faceHoverColorButton)
        self._hoverLayout.addRow('Edge Hover Color', self._edgeHoverColorButton)
        self._hoverLayout.addRow('Edge Hover Width', self._edgeHoverWidthSpinBox)

        self._hoverLayoutWrapper = QVBoxLayout()
        self._hoverLayoutWrapper.addLayout(self._hoverLayout)
        self._hoverLayoutWrapper.addStretch()

        self._formatLayoutWrapper = QHBoxLayout()
        self._formatLayoutWrapper.addLayout(self._formatLayout)
        self._formatLayoutWrapper.addLayout(self._hoverLayoutWrapper)

        self.setLayout(self._formatLayoutWrapper)

        # default settings
        self.setFormat(AxisRegion().getFormat())
    
    def getFormat(self):
        fmt = {}

        faceColor = self._faceColorButton.color()
        edgeColor = self._edgeColorButton.color()
        edgeWidth = self._edgeWidthSpinBox.value()
        fontSize = self._fontSizeSpinBox.value()
        fontColor = self._fontColorButton.color()
        faceHoverColor = self._faceHoverColorButton.color()
        edgeHoverColor = self._edgeHoverColorButton.color()
        edgeHoverWidth = self._edgeHoverWidthSpinBox.value()

        if faceColor is not None:
            fmt['facecolor'] = toColorStr(faceColor)
        if edgeColor is not None:
            fmt['edgecolor'] = toColorStr(edgeColor)
        if edgeWidth > 0:
            fmt['edgewidth'] = edgeWidth
        if fontSize > 0:
            fmt['fontsize'] = fontSize
        if fontColor is not None:
            fmt['fontcolor'] = toColorStr(fontColor)
        if faceHoverColor is not None:
            fmt['facehovercolor'] = toColorStr(faceHoverColor)
        if edgeHoverColor is not None:
            fmt['edgehovercolor'] = toColorStr(edgeHoverColor)
        if edgeHoverWidth > 0:
            fmt['edgehoverwidth'] = edgeHoverWidth

        return fmt

    def setFormat(self, fmt: dict):
        for key, value in fmt.items():
            key = key.lower()
            if key == 'facecolor':
                self._faceColorButton.setColor(toQColor(value))
            elif key == 'edgecolor':
                self._edgeColorButton.setColor(toQColor(value))
            elif key == 'edgewidth':
                self._edgeWidthSpinBox.setValue(value)
            elif key == 'facehovercolor':
                self._faceHoverColorButton.setColor(toQColor(value))
            elif key == 'edgehovercolor':
                self._edgeHoverColorButton.setColor(toQColor(value))
            elif key == 'edgehoverwidth':
                self._edgeHoverWidthSpinBox.setValue(value)
            elif key == 'fontsize':
                self._fontSizeSpinBox.setValue(value)
            elif key == 'fontcolor':
                self._fontColorButton.setColor(toQColor(value))


def editAxisRegion(region: AxisRegion = None, parent: QWidget = None, title: str = None) -> dict | None:
    if region is None:
        region = AxisRegion()
    
    panel = AxisRegionPanel()
    panel.layout().setContentsMargins(0, 0, 0, 0)
    panel.setState(region.getState())

    dlg = QDialog(parent)
    vbox = QVBoxLayout(dlg)
    vbox.addWidget(panel)

    btns = QDialogButtonBox()
    btns.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
    btns.accepted.connect(dlg.accept)
    btns.rejected.connect(dlg.reject)
    vbox.addWidget(btns)
    vbox.addStretch()

    if title is not None:
        dlg.setWindowTitle(title)
    dlg.setWindowModality(Qt.ApplicationModal)
    
    if dlg.exec() != QDialog.Accepted:
        return
    
    region.setState(panel.getState())

    return region.getState()


def formatAxisRegion(region: AxisRegion = None, parent: QWidget = None, title: str = None) -> dict | None:
    if region is None:
        region = AxisRegion()
    
    panel = AxisRegionFormatPanel()
    panel.layout().setContentsMargins(0, 0, 0, 0)
    panel.setFormat(region.getFormat())

    dlg = QDialog(parent)
    vbox = QVBoxLayout(dlg)
    vbox.addWidget(panel)

    btns = QDialogButtonBox()
    btns.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
    btns.accepted.connect(dlg.accept)
    btns.rejected.connect(dlg.reject)
    vbox.addWidget(btns)
    vbox.addStretch()

    if title is not None:
        dlg.setWindowTitle(title)
    dlg.setWindowModality(Qt.ApplicationModal)
    
    if dlg.exec() != QDialog.Accepted:
        return
    
    region.setFormat(panel.getFormat())

    return region.getFormat()


# def editMultipleAxisRegions(regions: list[AxisRegion], parent: QWidget = None, title: str = None):
#     panel = AxisRegionPanel()
#     panel.layout().setContentsMargins(0, 0, 0, 0)
#     states = [region.getState() for region in regions]
#     panel.setState(states[0])
#     for state in states[1:]:
#         if state['region'] != states[0]['region']:
#             self._minEdit.setText('')
#             self._maxEdit.setText('')
#         if state['movable'] != states[0]['movable']:
#             self._movableCheckBox.setCheckState(Qt.PartiallyChecked)
#         if state['text'] != states[0]['text']:
#             self._textEdit.setPlainText('')
#         for key in state['format']:
#             if key not in states[0]['format'] or state['format'][key] != states[0]['format'][key]:
#                 pass
#     panel.setState(shared_state)

#     dlg = QDialog(parent)
#     vbox = QVBoxLayout(dlg)
#     vbox.addWidget(panel)

#     btns = QDialogButtonBox()
#     btns.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
#     btns.accepted.connect(dlg.accept)
#     btns.rejected.connect(dlg.reject)
#     vbox.addWidget(btns)
#     vbox.addStretch()

#     if title is not None:
#         dlg.setWindowTitle(title)
#     dlg.setWindowModality(Qt.ApplicationModal)
    
#     if dlg.exec() != QDialog.Accepted:
#         return


def test_live():
    app = QApplication()

    ui = AxisRegionPanel()
    ui.show()

    app.exec()

if __name__ == '__main__':
    test_live()